# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for indexing electron diffraction spot patterns."""

from collections import namedtuple
from itertools import combinations
from operator import attrgetter
import warnings

from dask.diagnostics import ProgressBar
from numba import njit, prange
import numpy as np
from orix.crystal_map import CrystalMap, PhaseList
from orix.quaternion import Rotation
import psutil
import scipy

from pyxem.utils.diffraction import _cart2polar
from pyxem.utils.vectors import get_rotation_matrix_between_vectors
from pyxem.utils.vectors import get_angle_cartesian
from pyxem.utils.cuda_utils import (
    is_cupy_array,
    get_array_module,
    _correlate_polar_image_to_library_gpu,
    TPB,
)
from pyxem.utils._dask import _get_dask_array
from pyxem.utils.polar_transform_utils import (
    _cartesian_positions_to_polar,
    get_polar_pattern_shape,
    image_to_polar,
    get_template_polar_coordinates,
    _warp_polar_custom,
)
from diffpy.structure import Atom, Structure, Lattice
from orix.crystal_map import Phase

try:
    import cupy as cp

    CUPY_INSTALLED = True
    import cupyx.scipy as spgpu
except ImportError:
    CUPY_INSTALLED = False


# container for OrientationResults
OrientationResult = namedtuple(
    "OrientationResult",
    "phase_index rotation_matrix match_rate error_hkls total_error scale center_x center_y".split(),
)


def get_nth_best_solution(
    single_match_result, mode, rank=0, key="match_rate", descending=True
):
    """Get the nth best solution by match_rate from a pool of solutions

    Parameters
    ----------
    single_match_result : VectorMatchingResults, TemplateMatchingResults
        Pool of solutions from the vector matching algorithm
    mode : str
        'vector' or 'template'
    rank : int
        The rank of the solution, i.e. rank=2 returns the third best solution
    key : str
        The key to sort the solutions by, default = match_rate
    descending : bool
        Rank the keys from large to small

    Returns
    -------
    VectorMatching:
        best_fit : `OrientationResult`
            Parameters for the best fitting orientation
            Library Number, rotation_matrix, match_rate, error_hkls, total_error
    TemplateMatching: np.array
            Parameters for the best fitting orientation
            Library Number , [z, x, z], Correlation Score
    """
    if mode == "vector":
        try:
            best_fit = sorted(
                single_match_result[0].tolist(), key=attrgetter(key), reverse=descending
            )[rank]
        except AttributeError:
            best_fit = sorted(
                single_match_result.tolist(), key=attrgetter(key), reverse=descending
            )[rank]
    if mode == "template":
        srt_idx = np.argsort(single_match_result[:, 2])[::-1][rank]
        best_fit = single_match_result[srt_idx]

    return best_fit


def index_magnitudes(z, simulation, tolerance):
    """Assigns hkl indices to peaks in the diffraction profile.

    Parameters
    ----------
    simulation : DiffractionProfileSimulation
        Simulation of the diffraction profile.
    tolerance : float
        The n orientations with the highest correlation values are returned.

    Returns
    -------
    indexation : np.array()
        indexation results.

    """
    mags = z
    sim_mags = np.array(simulation.magnitudes)
    sim_hkls = np.array(simulation.hkls)
    indexation = np.zeros(len(mags), dtype=object)

    for i in np.arange(len(mags)):
        diff = np.absolute((sim_mags - mags.data[i]) / mags.data[i] * 100)

        hkls = sim_hkls[np.where(diff < tolerance)]
        diffs = diff[np.where(diff < tolerance)]

        indices = np.array((hkls, diffs))
        indexation[i] = np.array((mags.data[i], indices), dtype=object)

    return indexation


def _choose_peak_ids(peaks, n_peaks_to_index):
    """Choose `n_peaks_to_index` indices from `peaks`.

    This implementation sorts by angle and then picks every
    len(peaks)/n_peaks_to_index element to get an even distribution of angles.

    Parameters
    ----------
    peaks : array_like
        Array of peak positions.
    n_peaks_to_index : int
        Number of indices to return.

    Returns
    -------
    peak_ids : numpy.array
        Array of indices of the chosen peaks.
    """
    r, angles = _cart2polar(peaks[:, 0], peaks[:, 1])
    return angles.argsort()[
        np.linspace(0, angles.shape[0] - 1, n_peaks_to_index, dtype=int)
    ]


def match_vectors(
    peaks, library, mag_tol, angle_tol, index_error_tol, n_peaks_to_index, n_best
):
    # TODO: Sort peaks by intensity or SNR
    """Assigns hkl indices to pairs of diffraction vectors.

    Parameters
    ----------
    peaks : np.array()
        The experimentally measured diffraction vectors, associated with a
        particular probe position, to be indexed. In Cartesian coordinates.
    library : VectorLibrary
        Library of reciprocal space vectors to be matched to the vectors.
    mag_tol : float
        Max allowed magnitude difference when comparing vectors.
    angle_tol : float
        Max allowed angle difference in radians when comparing vector pairs.
    index_error_tol : float
        Max allowed error in peak indexation for classifying it as indexed,
        calculated as :math:`|hkl_calculated - round(hkl_calculated)|`.
    n_peaks_to_index : int
        The maximum number of peak to index.
    n_best : int
        The maximum number of good solutions to be retained for each phase.

    Returns
    -------
    indexation : np.array()
        A numpy array containing the indexation results, each result consisting of 5 entries:
            [phase index, rotation matrix, match rate, error hkls, total error]

    """
    if peaks.shape == (1,) and peaks.dtype == object:
        peaks = peaks[0]

    # Assign empty array to hold indexation results. The n_best best results
    # from each phase is returned.
    top_matches = np.empty(len(library) * n_best, dtype="object")
    res_rhkls = []

    # Iterate over phases in DiffractionVectorLibrary and perform indexation
    # on each phase, storing the best results in top_matches.
    for phase_index, (phase, structure) in enumerate(
        zip(library.values(), library.structures)
    ):
        solutions = []
        lattice_recip = structure.lattice.reciprocal()
        phase_indices = phase["indices"]
        phase_measurements = phase["measurements"]

        if peaks.shape[0] < 2:  # pragma: no cover
            continue

        # Choose up to n_peaks_to_index unindexed peaks to be paired in all
        # combinations.
        # TODO: Matching can be done iteratively where successfully indexed
        #       peaks are removed after each iteration. This can possibly
        #       handle overlapping patterns.
        # unindexed_peak_ids = range(min(peaks.shape[0], n_peaks_to_index))
        # TODO: Better choice of peaks (longest, highest SNR?)
        # TODO: Inline after choosing the best, and possibly require external sorting (if using sorted)?
        unindexed_peak_ids = _choose_peak_ids(peaks, n_peaks_to_index)

        # Find possible solutions for each pair of peaks.
        for vector_pair_index, peak_pair_indices in enumerate(
            list(combinations(unindexed_peak_ids, 2))
        ):
            # Consider a pair of experimental scattering vectors.
            q1, q2 = peaks[peak_pair_indices, :]
            q1_len, q2_len = np.linalg.norm(q1), np.linalg.norm(q2)

            # Ensure q1 is longer than q2 for consistent order.
            if q1_len < q2_len:  # pragma: no cover
                q1, q2 = q2, q1
                q1_len, q2_len = q2_len, q1_len

            # Calculate the angle between experimental scattering vectors.
            angle = get_angle_cartesian(q1, q2)

            # Get library indices for hkls matching peaks within tolerances.
            # TODO: phase are object arrays. Test performance of direct float arrays
            tolerance_mask = np.abs(phase_measurements[:, 0] - q1_len) < mag_tol
            tolerance_mask[tolerance_mask] &= (
                np.abs(phase_measurements[tolerance_mask, 1] - q2_len) < mag_tol
            )
            tolerance_mask[tolerance_mask] &= (
                np.abs(phase_measurements[tolerance_mask, 2] - angle) < angle_tol
            )

            # Iterate over matched library vectors determining the error in the
            # associated indexation.
            if np.count_nonzero(tolerance_mask) == 0:  # pragma: no cover
                continue

            # Reference vectors are cartesian coordinates of hkls
            reference_vectors = lattice_recip.cartesian(phase_indices[tolerance_mask])

            # Rotation from experimental to reference frame
            rotations = get_rotation_matrix_between_vectors(
                q1, q2, reference_vectors[:, 0], reference_vectors[:, 1]
            )

            # Index the peaks by rotating them to the reference coordinate
            # system. Use rotation directly since it is multiplied from the
            # right. Einsum gives list of peaks.dot(rotation).
            hklss = lattice_recip.fractional(np.einsum("ijk,lk->ilj", rotations, peaks))

            # Evaluate error of peak hkl indexation
            rhklss = np.rint(hklss)
            ehklss = np.abs(hklss - rhklss)
            valid_peak_mask = np.max(ehklss, axis=-1) < index_error_tol
            valid_peak_counts = np.count_nonzero(valid_peak_mask, axis=-1)
            error_means = ehklss.mean(axis=(1, 2))

            num_peaks = len(peaks)
            match_rates = (valid_peak_counts * (1 / num_peaks)) if num_peaks else 0

            possible_solution_mask = match_rates > 0
            solutions += [
                OrientationResult(
                    phase_index=phase_index,
                    rotation_matrix=R,
                    match_rate=match_rate,
                    error_hkls=ehkls,
                    total_error=error_mean,
                    scale=1.0,
                    center_x=0.0,
                    center_y=0.0,
                )
                for R, match_rate, ehkls, error_mean in zip(
                    rotations[possible_solution_mask],
                    match_rates[possible_solution_mask],
                    ehklss[possible_solution_mask],
                    error_means[possible_solution_mask],
                )
            ]

            res_rhkls += rhklss[possible_solution_mask].tolist()

        n_solutions = min(n_best, len(solutions))

        i = phase_index * n_best  # starting index in unfolded array

        if n_solutions > 0:
            top_n = sorted(solutions, key=attrgetter("match_rate"), reverse=True)[
                :n_solutions
            ]

            # Put the top n ranked solutions in the output array
            top_matches[i : i + n_solutions] = top_n

        if n_solutions < n_best:
            # Fill with dummy values
            top_matches[i + n_solutions : i + n_best] = [
                OrientationResult(
                    phase_index=0,
                    rotation_matrix=np.identity(3),
                    match_rate=0.0,
                    error_hkls=np.array([]),
                    total_error=1.0,
                    scale=1.0,
                    center_x=0.0,
                    center_y=0.0,
                )
                for x in range(n_best - n_solutions)
            ]

    # Because of a bug in numpy (https://github.com/numpy/numpy/issues/7453),
    # triggered by the way HyperSpy reads results (np.asarray(res), which fails
    # when the two tuple values have the same first dimension), we cannot
    # return a tuple directly, but instead have to format the result as an
    # array ourselves.
    res = np.empty(2, dtype=object)
    res[0] = top_matches
    res[1] = np.asarray(res_rhkls)
    return res


def _simulations_to_arrays(simulations, max_radius=None):
    """
    Convert simulation results to arrays of diffraction spots

    Parameters
    ----------
    simulations : list
        list of diffsims.sims.diffraction_simulation.DiffractionSimulation
        objects
    max_radius : float
        limit to g-vector length in pixel coordinates

    Returns
    -------
    positions : numpy.ndarray (N, 2, R)
        An array containing all (x,y) coordinates of reflections of N templates. R represents
        the maximum number of reflections; templates containing fewer
        reflections are padded with 0's at the end. In pixel units.
    intensities : numpy.ndarray (N, R)
        An array containing all intensities of reflections of N templates
    """
    num_spots = [i.intensities.shape[0] for i in simulations]
    max_spots = max(num_spots)
    positions = np.zeros((len(simulations), 2, max_spots), dtype=np.float64)
    intensities = np.zeros((len(simulations), max_spots), dtype=np.float64)
    for i, j in enumerate(simulations):
        x = j.calibrated_coordinates[:, 0]
        y = j.calibrated_coordinates[:, 1]
        intensity = j.intensities
        if max_radius is not None:
            condition = x**2 + y**2 < max_radius**2
            x = x[condition]
            y = y[condition]
            intensity = intensity[condition]
        positions[i, 0, : x.shape[0]] = x
        positions[i, 1, : y.shape[0]] = y
        intensities[i, : intensity.shape[0]] = intensity
    return positions, intensities


def _match_polar_to_polar_template(
    polar_image,
    r_template,
    theta_template,
    intensities,
):
    """
    Correlate a single polar template to a single polar image

    The template spots are shifted along the azimuthal axis by 1 pixel increments.
    A simple correlation index is calculated at each position.

    Parameters
    ----------
    polar_image : 2D ndarray
        the polar image
    r_template : 1D ndarray
        r coordinates of diffraction spots in template
    theta_template : 1D ndarray
        theta coordinates of diffraction spots in template
    intensities : 1D ndarray
        intensities of diffraction spots in template

    Returns
    -------
    correlation : 1D ndarray
        correlation index at each in-plane angle position
    """
    dispatcher = get_array_module(polar_image)
    sli = polar_image[:, r_template]
    rows, column_indices = dispatcher.ogrid[: sli.shape[0], : sli.shape[1]]
    rows = dispatcher.mod(rows + theta_template[None, :], polar_image.shape[0])
    extr = sli[rows, column_indices].astype(intensities.dtype)
    correlation = dispatcher.dot(extr, intensities)
    return correlation


@njit(parallel=True, nogil=True)
def _match_polar_to_polar_library_cpu(
    polar_image,
    r_templates,
    theta_templates,
    intensities_templates,
):
    """
    Correlates a polar pattern to all polar templates on CPU

    Parameters
    ----------
    polar_image : 2D numpy.ndarray
        The image converted to polar coordinates
    r_templates : 2D numpy.ndarray
        r-coordinates of diffraction spots in templates.
    theta_templates : 2D numpy ndarray
        theta-coordinates of diffraction spots in templates.
    intensities_templates : 2D numpy.ndarray
        intensities of the spots in each template

    Returns
    -------
    best_in_plane_shift : (N) 1D numpy.ndarray
        Shift for all templates that yields best correlation
    best_in_plane_corr : (N) 1D numpy.ndarray
        Correlation at best match for each template
    best_in_plane_shift_m : (N) 1D numpy.ndarray
        Shift for all mirrored templates that yields best correlation
    best_in_plane_corr_m : (N) 1D numpy.ndarray
        Correlation at best match for each mirrored template

    Notes
    -----
    The dimensions of r_templates and theta_templates should be (N, R) where
    N is the number of templates and R the number of spots in the template
    with the maximum number of spots
    """
    N = r_templates.shape[0]
    R = r_templates.shape[1]
    n_shifts = polar_image.shape[0]
    best_in_plane_shift = np.empty(N, dtype=np.int32)
    best_in_plane_shift_m = np.empty(N, dtype=np.int32)
    best_in_plane_corr = np.empty(N, dtype=polar_image.dtype)
    best_in_plane_corr_m = np.empty(N, dtype=polar_image.dtype)

    for template in prange(N):
        inplane_cor = np.zeros(n_shifts)
        inplane_cor_m = np.zeros(n_shifts)
        for spot in range(R):
            rsp = r_templates[template, spot]
            if rsp == 0:
                break
            tsp = theta_templates[template, spot]
            isp = intensities_templates[template, spot]
            split = n_shifts - tsp
            column = polar_image[:, rsp] * isp
            inplane_cor[:split] += column[tsp:]
            inplane_cor[split:] += column[:tsp]
            inplane_cor_m[:tsp] += column[split:]
            inplane_cor_m[tsp:] += column[:split]

        best_shift = np.argmax(inplane_cor)
        best_shift_m = np.argmax(inplane_cor_m)
        best_in_plane_shift[template] = best_shift
        best_in_plane_shift_m[template] = best_shift_m
        best_in_plane_corr[template] = inplane_cor[best_shift]
        best_in_plane_corr_m[template] = inplane_cor_m[best_shift_m]

    return (
        best_in_plane_shift,
        best_in_plane_corr,
        best_in_plane_shift_m,
        best_in_plane_corr_m,
    )


def _match_polar_to_polar_library_gpu(
    polar_image,
    r_templates,
    theta_templates,
    intensities_templates,
):
    """
    Correlates a polar pattern to all polar templates on GPU

    Parameters
    ----------
    polar_image : 2D cupy.ndarray
        The image converted to polar coordinates
    r_templates : 2D cupy.ndarray
        r-coordinates of diffraction spots in templates.
    theta_templates : 2D cupy ndarray
        theta-coordinates of diffraction spots in templates.
    intensities_templates : 2D cupy.ndarray
        intensities of the spots in each template

    Returns
    -------
    best_in_plane_shift : (N) 1D cupy.ndarray
        Shift for all templates that yields best correlation
    best_in_plane_corr : (N) 1D cupy.ndarray
        Correlation at best match for each template
    best_in_plane_shift_m : (N) 1D cupy.ndarray
        Shift for all mirrored templates that yields best correlation
    best_in_plane_corr_m : (N) 1D cupy.ndarray
        Correlation at best match for each mirrored template

    Notes
    -----
    The dimensions of r_templates and theta_templates should be (N, R) where
    N is the number of templates and R the number of spots in the template
    with the maximum number of spots
    """
    correlation = cp.empty(
        (r_templates.shape[0], polar_image.shape[0]), dtype=cp.float32
    )
    correlation_m = cp.empty(
        (r_templates.shape[0], polar_image.shape[0]), dtype=cp.float32
    )
    threadsperblock = (1, TPB)
    blockspergrid = (r_templates.shape[0], int(np.ceil(polar_image.shape[0] / TPB)))
    _correlate_polar_image_to_library_gpu[blockspergrid, threadsperblock](
        polar_image,
        r_templates,
        theta_templates,
        intensities_templates,
        correlation,
        correlation_m,
    )
    best_in_plane_shift = cp.argmax(correlation, axis=1).astype(np.int32)
    best_in_plane_shift_m = cp.argmax(correlation_m, axis=1).astype(np.int32)
    rows = cp.arange(correlation.shape[0], dtype=np.int32)
    best_in_plane_corr = correlation[rows, best_in_plane_shift]
    best_in_plane_corr_m = correlation_m[rows, best_in_plane_shift_m]
    return (
        best_in_plane_shift,
        best_in_plane_corr,
        best_in_plane_shift_m,
        best_in_plane_corr_m,
    )


def _get_row_norms(array):
    """Get the norm of all rows in a 2D array"""
    norms = ((array**2).sum(axis=1)) ** 0.5
    return norms


def _norm_rows(array):
    """Normalize all the rows in a 2D array"""
    norms = _get_row_norms(array)
    return array / norms[:, None]


def _get_integrated_polar_templates(
    r_max, r_templates, intensities_templates, normalize_templates
):
    """
    Get an azimuthally integrated representation of the templates.

    Parameters
    ----------
    r_max : float
        maximum radial distance to consider in pixel units. Typically the
        radial width of the polar images.
    r_templates : 2D numpy or cupy ndarray
        r-coordinate of all spots in the templates. Of shape (N, R) where
        N is the number of templates and R is the number of spots in the
        template with the maximum number of spots
    intensities_templates : 2D numpy or cupy ndarray
        intensities in all spots of the templates. Of shape (N, R)
    normalize_templates : bool
        Whether to normalize the integrated templates

    Returns
    -------
    integrated_templates : 2D numpy or cupy ndarray
        Templates integrated over the azimuthal axis of shape (N, r_max)
    """
    dispatcher = get_array_module(intensities_templates)
    data = intensities_templates.ravel()
    columns = r_templates.ravel()
    rows = dispatcher.arange(r_templates.shape[0]).repeat(r_templates.shape[1])
    out_shape = (r_templates.shape[0], r_max)
    if is_cupy_array(intensities_templates):
        integrated_templates = spgpu.sparse.coo_matrix(
            (data, (rows, columns)), shape=out_shape
        ).toarray()
    else:
        integrated_templates = scipy.sparse.coo_matrix(
            (data, (rows, columns)), shape=out_shape
        ).toarray()
    if normalize_templates:
        integrated_templates = _norm_rows(integrated_templates)
    return integrated_templates


def _match_library_to_polar_fast(polar_sum, integrated_templates):
    """
    Compare a polar image to azimuthally integrated templates

    Parameters
    ----------
    polar_sum : 1D numpy array or cupy array
        the image in polar coordinates integrated along the azimuthal axis
        (shape = (r_max,))
    integrated_templates : 2D numpy array or cupy array
        azimuthally integrated templates of shape (N, r_max) with N
        the number of templates and r_max the width of the polar image

    Returns
    -------
    correlations : 1D numpy array or cupy array
        the correlation between the integrated image and the integrated
        templates. (shape = (N,))
    """
    return (integrated_templates * polar_sum).sum(axis=1)


def _prepare_image_and_templates(
    image,
    simulations,
    delta_r,
    delta_theta,
    max_r,
    intensity_transform_function,
    find_direct_beam,
    direct_beam_position,
    normalize_image,
    normalize_templates,
):
    """
    Prepare a single cartesian coordinate image and a template library for comparison

    Parameters
    ----------
    image : 2D np or cp ndarray
        The diffraction pattern in cartesian coordinates
    simulations : list
        list of diffsims.sims.diffraction_simulation.DiffractionSimulation
    delta_r : float
        sampling interval for the r coordinate in the polar image in pixels
    delta_theta : float
        sampling interval for the theta coordinate in the polar image in degrees
    max_r : float
        maximum radius to consider in polar conversion, in pixels
    intensity_transform_function : Callable
        function to apply to both the image and template intensities. Must
        accept any dimensional numpy array as input and preferably operate
        independently on individual elements
    find_direct_beam : bool
        whether to refine the direct beam position in the image polar conversion
    direct_beam_position : 2-tuple of floats
        the (x, y) position of the direct beam in the image to override any
        defaults
    normalize_image : bool
        Whether to normalize the image
    normalize_templates : bool
        Whether to normalize the template intensities

    Returns
    -------
    polar_image : 2D np or cp ndarray
        The image in polar coordinates
    r : 2D np or cp ndarray
        The r coordinates in the polar image corresponding to template spots.
        shape = (N, R) with N the number of templates and R the number of spots
        contained by the template with the most spots
    theta : 2D np or cp ndarray
        The theta coordinates in the polar image corresponding to template spots.
        shape = (N, R) with N the number of templates and R the number of spots
        contained by the template with the most spots
    intensities :  2D np or cp ndarray
        The intensities corresponding to template spots.
        shape = (N, R) with N the number of templates and R the number of spots
        contained by the template with the most spots
    """
    polar_image = image_to_polar(
        image,
        delta_r,
        delta_theta,
        max_r=max_r,
        find_direct_beam=find_direct_beam,
        direct_beam_position=direct_beam_position,
    )
    dispatcher = get_array_module(polar_image)
    max_radius = polar_image.shape[1] * delta_r
    positions, intensities = _simulations_to_arrays(simulations, max_radius=max_radius)
    r, theta = _cartesian_positions_to_polar(
        positions[:, 0], positions[:, 1], delta_r, delta_theta
    )
    condition = r >= polar_image.shape[1]
    # we don't set r to 0 because it may quit the loop in the matching early
    r[condition] = polar_image.shape[1] - 1
    theta[condition] = 0
    intensities[condition] = 0.0
    if is_cupy_array(polar_image):
        # send data to GPU
        r = cp.asarray(r)
        theta = cp.asarray(theta)
        intensities = cp.asarray(intensities)
    if intensity_transform_function is not None:
        intensities = intensity_transform_function(intensities)
        polar_image = intensity_transform_function(polar_image)
    if normalize_image:
        polar_image = polar_image / dispatcher.linalg.norm(polar_image)
    if normalize_templates:
        intensities = _norm_rows(intensities)
    return (polar_image, r, theta, intensities)


def _mixed_matching_lib_to_polar(
    polar_image,
    integrated_templates,
    r_templates,
    theta_templates,
    intensities_templates,
    n_keep,
    frac_keep,
    n_best,
    transpose=False,
):
    """
    Match a polar image to a filtered subset of polar templates

    First does a fast matching basted on azimuthally integrated templates
    Then it takes the (1-fraction)*100% of patterns to do a full indexation on.
    Return the first n_best answers.

    Parameters
    ----------
    polar_image : 2D ndarray
        image in polar coordinates
    integrated_templates : 2D ndarray, (N, r_max)
        azimuthally integrated templates
    r_templates : 2D ndarray, (N, R)
        r coordinates of diffraction spots in all N templates
    theta_templates : 2D ndarray, (N, R)
        theta coordinates of diffraction spots in all N templates
    intensities_templates : 2D ndarray, (N, R)
        intensities of diffraction spots in all N templates
    frac_keep : float
        fraction of templates to pass on to the full indexation
    n_keep : float
        number of templates to pass to the full indexation
    n_best : int
        number of solutions to return in decending order of fit

    Return
    ------
    answer : 2D numpy array, (n_best, 4)
        in the colums are returned (template index, correlation, in-plane angle, factor)
        of the best fitting template, where factor is 1 if the direct template is
        matched and -1 if the mirror template is matched
    """
    if transpose:
        polar_image = polar_image.T
    polar_image = np.nan_to_num(polar_image)
    dispatcher = get_array_module(polar_image)
    # remove templates we don't care about with a fast match
    (
        template_indexes,
        r_templates,
        theta_templates,
        intensities_templates,
    ) = _prefilter_templates(
        polar_image,
        r_templates,
        theta_templates,
        intensities_templates,
        integrated_templates,
        frac_keep,
        n_keep,
    )
    # get a full match on the filtered data - we must branch for CPU/GPU
    (
        best_in_plane_shift,
        best_in_plane_corr,
        best_in_plane_shift_m,
        best_in_plane_corr_m,
    ) = _get_full_correlations(
        polar_image,
        r_templates,
        theta_templates,
        intensities_templates,
    )
    # compare positive and negative templates and combine
    positive_is_best = best_in_plane_corr >= best_in_plane_corr_m
    negative_is_best = ~positive_is_best
    # multiplication method is faster than dispatcher.choose
    best_sign = positive_is_best * 1 + negative_is_best * (-1)
    best_cors = (
        positive_is_best * best_in_plane_corr + negative_is_best * best_in_plane_corr_m
    )
    best_angles = (
        positive_is_best * best_in_plane_shift
        + negative_is_best * best_in_plane_shift_m
    )
    if n_best >= best_cors.shape[0]:
        n_best = best_cors.shape[0]
    if n_best < 1:
        nbest = 1
    answer = dispatcher.empty((n_best, 4), dtype=polar_image.dtype)
    if n_best == 1:
        max_index_filter = dispatcher.argmax(best_cors)
        max_cor = best_cors[max_index_filter]
        max_angle = best_angles[max_index_filter]
        max_index = template_indexes[max_index_filter]
        max_sign = best_sign[max_index_filter]
        answer[0] = dispatcher.array((max_index, max_cor, max_angle, max_sign))
    else:
        # a partial sort
        indices_nbest = dispatcher.argpartition(-best_cors, n_best - 1)[:n_best]
        nbest_cors = best_cors[indices_nbest]
        # a full sort on this subset
        indices_sorted = dispatcher.argsort(-nbest_cors)
        n_best_indices = indices_nbest[indices_sorted]
        answer[:, 0] = template_indexes[n_best_indices]
        answer[:, 1] = best_cors[n_best_indices]
        answer[:, 2] = best_angles[n_best_indices]
        answer[:, 3] = best_sign[n_best_indices]
    return answer


def _index_chunk(
    images,
    center,
    max_radius,
    output_shape,
    precision,
    integrated_templates,
    r_templates,
    theta_templates,
    intensities_templates,
    n_keep,
    frac_keep,
    n_best,
    norm_images,
    order=1,
):
    dispatcher = get_array_module(images)
    # prepare an empty results chunk
    indexation_result_chunk = dispatcher.empty(
        (images.shape[0], images.shape[1], n_best, 4),
        dtype=precision,
    )
    for index in np.ndindex(images.shape[:2]):
        polar_image = _warp_polar_custom(
            images[index],
            center,
            max_radius,
            output_shape,
            order=order,
            precision=precision,
        )
        if norm_images:
            polar_image = polar_image / dispatcher.linalg.norm(polar_image)
        indexation_result_chunk[index] = _mixed_matching_lib_to_polar(
            polar_image,
            integrated_templates,
            r_templates,
            theta_templates,
            intensities_templates,
            n_keep,
            frac_keep,
            n_best,
        )
    return indexation_result_chunk


def _index_chunk_gpu(images, *args, **kwargs):
    gpu_im = cp.asarray(images)
    indexed_chunk = _index_chunk(gpu_im, *args, **kwargs)
    return cp.asnumpy(indexed_chunk)


def get_in_plane_rotation_correlation(
    image,
    simulation,
    intensity_transform_function=None,
    delta_r=1,
    delta_theta=1,
    max_r=None,
    find_direct_beam=False,
    direct_beam_position=None,
    normalize_image=True,
    normalize_template=True,
):
    """
    Correlate a single image and simulation over the in-plane rotation angle

    Parameters
    ----------
    image : 2D numpy.ndarray
        The image of the diffraction pattern
    simulation : diffsims.sims.diffraction_simulation.DiffractionSimulation
        The diffraction pattern simulation
    intensity_transform_function : Callable, optional
        Function to apply to both image and template intensities on an
        element by element basis prior to comparison
    delta_r : float, optional
        The sampling interval of the radial coordinate in pixels
    delta_theta : float, optional
        The sampling interval of the azimuthal coordinate in degrees
    max_r : float, optional
        Maximum radius to consider in pixel units. By default it is the
        distance from the center of the image to a corner
    find_direct_beam : bool, optional
        Whether to optimize the direct beam, otherwise the center of the image
        is chosen
    direct_beam_position : 2-tuple of floats, optional
        (x, y) coordinate of the direct beam in pixel units. Overrides other
        settings for finding the direct beam
    normalize_image : bool, optional
        normalize the image to calculate the correlation coefficient
    normalize_template : bool, optional
        normalize the template to calculate the correlation coefficient

    Returns
    -------
    angle_array : 1D np.ndarray
        The in-plane angles at which the correlation is calculated in degrees
    correlation_array : 1D np.ndarray
        The correlation corresponding to these angles
    """
    polar_image = image_to_polar(
        image,
        delta_r,
        delta_theta,
        max_r=max_r,
        find_direct_beam=find_direct_beam,
        direct_beam_position=direct_beam_position,
    )
    r, theta, intensity = get_template_polar_coordinates(
        simulation,
        in_plane_angle=0.0,
        delta_r=delta_r,
        delta_theta=delta_theta,
        max_r=polar_image.shape[1],
    )
    if is_cupy_array(polar_image):
        dispatcher = cp
        r = cp.asarray(r)
        theta = cp.asarray(theta)
        intensity = cp.asarray(intensity)
    else:
        dispatcher = np
    if intensity_transform_function is not None:
        intensity = intensity_transform_function(intensity)
        polar_image = intensity_transform_function(polar_image)
    if normalize_image:
        polar_image = polar_image / dispatcher.linalg.norm(polar_image)
    if normalize_template:
        intensity = intensity / dispatcher.linalg.norm(intensity)
    correlation_array = _match_polar_to_polar_template(
        polar_image,
        r,
        theta,
        intensity,
    )
    angle_array = dispatcher.arange(correlation_array.shape[0]) * delta_theta
    return angle_array, correlation_array


def _get_fast_correlation_index(
    polar_image,
    r,
    intensities,
    normalize_image,
    normalize_templates,
):
    dispatcher = get_array_module(polar_image)
    integrated_polar = polar_image.sum(axis=0)
    rrr = dispatcher.arange(integrated_polar.shape[0]) / integrated_polar.shape[0]
    integrated_polar = integrated_polar * rrr
    integrated_templates = _get_integrated_polar_templates(
        integrated_polar.shape[0],
        r,
        intensities,
        normalize_templates,
    )
    if normalize_image:
        integrated_polar = integrated_polar / np.linalg.norm(integrated_polar)
    correlations = _match_library_to_polar_fast(
        integrated_polar,
        integrated_templates,
    )
    return correlations


def correlate_library_to_pattern_fast(
    image,
    simulations,
    delta_r=1,
    delta_theta=1,
    max_r=None,
    intensity_transform_function=None,
    find_direct_beam=False,
    direct_beam_position=None,
    normalize_image=True,
    normalize_templates=True,
):
    """
    Get the correlation between azimuthally integrated templates and patterns

    Parameters
    ----------
    image : 2D numpy.ndarray
        The image of the diffraction pattern
    simulations : list of diffsims.sims.diffraction_simulation.DiffractionSimulation
        The diffraction pattern simulation
    delta_r : float, optional
        The sampling interval of the radial coordinate in pixels
    delta_theta : float, optional
        The sampling interval of the azimuthal coordinate in degrees
    max_r : float, optional
        Maximum radius to consider in pixel units. By default it is the
        distance from the center of the image to a corner
    intensity_transform_function : Callable, optional
        Function to apply to both image and template intensities on an
        element by element basis prior to comparison
    find_direct_beam : bool, optional
        Whether to optimize the direct beam, otherwise the center of the image
        is chosen
    direct_beam_position : 2-tuple of floats, optional
        (x, y) coordinate of the direct beam in pixel units. Overrides other
        settings for finding the direct beam
    normalize_image : bool, optional
        normalize the image to calculate the correlation coefficient
    normalize_templates : bool, optional
        normalize the templates to calculate the correlation coefficient

    Returns
    -------
    correlations : 1D numpy.ndarray
        correlation between azimuthaly integrated template and each azimuthally integrated template

    Notes
    -----
    Mirrored templates have identical azimuthally integrated representations,
    so this only has to be done on the positive euler angle templates (0, Phi, phi2)
    """
    polar_image, r, theta, intensities = _prepare_image_and_templates(
        image,
        simulations,
        delta_r,
        delta_theta,
        max_r,
        intensity_transform_function,
        find_direct_beam,
        direct_beam_position,
        False,  # it is not necessary to normalize these here
        False,
    )
    return _get_fast_correlation_index(
        polar_image, r, intensities, normalize_image, normalize_templates
    )


def _get_max_n(N, n_keep, frac_keep):
    """
    Determine the number of templates to allow through
    """
    max_keep = N
    if frac_keep is not None:
        max_keep = max(round(frac_keep * N), 1)
    # n_keep overrides fraction
    if n_keep is not None:
        max_keep = max(n_keep, 1)
    return int(min(max_keep, N))


def _prefilter_templates(
    polar_image,
    r,
    theta,
    intensities,
    integrated_templates,
    frac_keep,
    n_keep,
):
    """
    Pre-filter the templates to reduce the number of templates to do a full match on

    Parameters
    ----------
    polar_image: 2D numpy.ndarray
        The image converted to polar coordinates in the form (theta, r)
    r:
    theta
    intensities
    integrated_templates
    frac_keep
    n_keep

    Returns
    -------

    """
    dispatcher = get_array_module(polar_image)
    max_keep = _get_max_n(r.shape[0], n_keep, frac_keep)
    template_indexes = dispatcher.arange(r.shape[0], dtype=np.int32)
    if max_keep != r.shape[0]:
        polar_sum = polar_image.sum(axis=0)
        rrr = dispatcher.arange(polar_sum.shape[0]) / polar_sum.shape[0]
        polar_sum = polar_sum * rrr
        correlations_fast = _match_library_to_polar_fast(
            polar_sum,
            integrated_templates,
        )
        sorted_cor_indx = dispatcher.argsort(-correlations_fast)[:max_keep]
        r = r[sorted_cor_indx]
        theta = theta[sorted_cor_indx]
        intensities = intensities[sorted_cor_indx]
        template_indexes = template_indexes[sorted_cor_indx]
    return template_indexes, r, theta, intensities


def _get_full_correlations(
    polar_image,
    r,
    theta,
    intensities,
):
    # get a full match on the filtered data - we must branch for CPU/GPU
    if is_cupy_array(polar_image):
        f = _match_polar_to_polar_library_gpu
    else:
        f = _match_polar_to_polar_library_cpu
    return f(polar_image, r, theta, intensities)


def correlate_library_to_pattern(
    image,
    simulations,
    frac_keep=1.0,
    n_keep=None,
    delta_r=1.0,
    delta_theta=1.0,
    max_r=None,
    intensity_transform_function=None,
    find_direct_beam=False,
    direct_beam_position=None,
    normalize_image=True,
    normalize_templates=True,
):
    """
    Get the best angle and associated correlation values, as well as the correlation with the inverted templates

    Parameters
    ----------
    image : 2D numpy.ndarray
        The image of the diffraction pattern
    simulations : list of diffsims.sims.diffraction_simulation.DiffractionSimulation
        The diffraction pattern simulation
    frac_keep : float
        Fraction (between 0-1) of templates to do a full matching on. By default
        all patterns are fully matched.
    n_keep : int
        Number of templates to do a full matching on. When set frac_keep will be
        ignored
    delta_r : float, optional
        The sampling interval of the radial coordinate in pixels
    delta_theta : float, optional
        The sampling interval of the azimuthal coordinate in degrees
    max_r : float, optional
        Maximum radius to consider in pixel units. By default it is the
        distance from the center of the image to a corner
    intensity_transform_function : Callable, optional
        Function to apply to both image and template intensities on an
        element by element basis prior to comparison
    find_direct_beam : bool, optional
        Whether to optimize the direct beam, otherwise the center of the image
        is chosen
    direct_beam_position : 2-tuple of floats, optional
        (x, y) coordinate of the direct beam in pixel units. Overrides other
        settings for finding the direct beam
    normalize_image : bool, optional
        normalize the image to calculate the correlation coefficient
    normalize_templates : bool, optional
        normalize the templates to calculate the correlation coefficient

    Returns
    -------
    indexes : 1D numpy.ndarray
        indexes of templates on which a full calculation has been performed
    angles : 1D numpy.ndarray
        best fit in-plane angle for the top "keep" templates
    correlations : 1D numpy.ndarray
        best correlation for the top "keep" templates
    angles_mirrored : 1D numpy.ndarray
        best fit in-plane angle for the top mirrored "keep" templates
    correlations_mirrored : 1D numpy.ndarray
        best correlation for the top mirrored "keep" templates

    Notes
    -----
    Mirrored refers to the templates corresponding to the inverted orientations
    (0, -Phi, -phi/2)
    """
    polar_image, r, theta, intensities = _prepare_image_and_templates(
        image,
        simulations,
        delta_r,
        delta_theta,
        max_r,
        intensity_transform_function,
        find_direct_beam,
        direct_beam_position,
        normalize_image,
        normalize_templates,
    )
    integrated_templates = _get_integrated_polar_templates(
        polar_image.shape[1],
        r,
        intensities,
        normalize_templates,
    )
    indexes, r, theta, intensities = _prefilter_templates(
        polar_image, r, theta, intensities, integrated_templates, frac_keep, n_keep
    )
    angles, cor, angles_m, cor_m = _get_full_correlations(
        polar_image,
        r,
        theta,
        intensities,
    )
    return (
        indexes,
        (angles * delta_theta).astype(polar_image.dtype),
        cor,
        (angles_m * delta_theta).astype(polar_image.dtype),
        cor_m,
    )


def get_n_best_matches(
    image,
    simulations,
    n_best=1,
    frac_keep=1.0,
    n_keep=None,
    delta_r=1.0,
    delta_theta=1.0,
    max_r=None,
    intensity_transform_function=None,
    find_direct_beam=False,
    direct_beam_position=None,
    normalize_image=True,
    normalize_templates=True,
):
    """
    Get the n templates best matching an image in descending order

    Parameters
    ----------
    image : 2D numpy or cupy ndarray
        The image of the diffraction pattern
    simulations : list of diffsims.sims.diffraction_simulation.DiffractionSimulation
        The diffraction pattern simulation
    n_best : int, optional
        Number of best solutions to return, in order of descending match
    n_keep : int, optional
        Number of templates to do a full matching on
    frac_keep : float, optional
        Fraction (between 0-1) of templates to do a full matching on. When set
        n_keep will be ignored
    delta_r : float, optional
        The sampling interval of the radial coordinate in pixels
    delta_theta : float, optional
        The sampling interval of the azimuthal coordinate in degrees
    max_r : float, optional
        Maximum radius to consider in pixel units. By default it is the
        distance from the center of the image to a corner
    intensity_transform_function : Callable, optional
        Function to apply to both image and template intensities on an
        element by element basis prior to comparison
    find_direct_beam : bool, optional
        Whether to optimize the direct beam, otherwise the center of the image
        is chosen
    direct_beam_position : 2-tuple of floats, optional
        (x, y) coordinate of the direct beam in pixel units. Overrides other
        settings for finding the direct beam
    normalize_image : bool, optional
        normalize the image to calculate the correlation coefficient
    normalize_templates : bool, optional
        normalize the templates to calculate the correlation coefficient

    Returns
    -------
    indexes : 1D numpy or cupy ndarray
        indexes of best fit templates
    angles : 1D numpy or cupy ndarray
        corresponding best fit in-plane angles
    correlations : 1D numpy or cupy ndarray
        corresponding correlation values
    signs : 1D numpy or cupy ndarray
        1 if the positive template (0, Phi, phi2) is best matched, -1 if
        the negative template (0, -Phi, -phi2) is best matched
    """
    polar_image, r, theta, intensities = _prepare_image_and_templates(
        image,
        simulations,
        delta_r,
        delta_theta,
        max_r,
        intensity_transform_function,
        find_direct_beam,
        direct_beam_position,
        normalize_image,
        normalize_templates,
    )
    integrated_templates = _get_integrated_polar_templates(
        polar_image.shape[1], r, intensities, normalize_templates
    )
    answer = _mixed_matching_lib_to_polar(
        polar_image,
        integrated_templates,
        r,
        theta,
        intensities,
        n_keep,
        frac_keep,
        n_best,
    )
    indices = answer[:, 0].astype(np.int32)
    cors = answer[:, 1]
    angles = (answer[:, 2] * delta_theta).astype(polar_image.dtype)
    sign = answer[:, 3]
    return indices, angles, cors, sign


def index_dataset_with_template_rotation(
    signal,
    library,
    phases=None,
    n_best=1,
    frac_keep=1.0,
    n_keep=None,
    delta_r=1.0,
    delta_theta=1.0,
    max_r=None,
    intensity_transform_function=None,
    normalize_images=False,
    normalize_templates=True,
    chunks="auto",
    parallel_workers="auto",
    target="cpu",
    scheduler="threads",
    precision=np.float64,
):
    """
    Index a dataset with template_matching while simultaneously optimizing in-plane rotation angle of the templates

    Parameters
    ----------
    signal : hyperspy.signals.Signal2D
        The 4D-STEM dataset.
    library : diffsims.libraries.diffraction_library.DiffractionLibrary
        The library of simulated diffraction patterns.
    phases : list, optional
        Names of phases in the library to do an indexation for. By default this is
        all phases in the library.
    n_best : int, optional
        Number of best solutions to return, in order of descending match.
    frac_keep : float, optional
        Fraction (between 0-1) of templates to do a full matching on. By default
        all templates will be fully matched. See notes for details.
    n_keep : int, optional
        Number of templates to do a full matching on. Overrides frac_keep.
    delta_r : float, optional
        The sampling interval of the radial coordinate in pixels.
    delta_theta : float, optional
        The sampling interval of the azimuthal coordinate in degrees. This will
        determine the maximum accuracy of the in-plane rotation angle.
    max_r : float, optional
        Maximum radius to consider in pixel units. By default it is the
        distance from the center of the patterns to a corner of the image.
    intensity_transform_function : Callable, optional
        Function to apply to both image and template intensities on an
        element by element basis prior to comparison. Note that the function
        is performed on the CPU.
    normalize_images : bool, optional
        Normalize the images in the correlation coefficient calculation
    normalize_templates : bool, optional
        Normalize the templates in the correlation coefficient calculation
    chunks : string or 4-tuple, optional
        Internally the work is done on dask arrays and this parameter determines
        the chunking of the original dataset. If set to None then no
        re-chunking will happen if the dataset was loaded lazily. If set to
        "auto" then dask attempts to find the optimal chunk size.
    parallel_workers: int, optional
        The number of workers to use in parallel. If set to "auto", the number
        of physical cores will be used when using the CPU. For GPU calculations
        the workers is determined based on the VRAM capacity, but it is probably
        better to choose a lower number.
    target: string, optional
        Use "cpu" or "gpu". If "gpu" is selected, the majority of the calculation
        intensive work will be performed on the CUDA enabled GPU. Fails if no
        such hardware is available.
    scheduler: string
        The scheduler used by dask to compute the result. "processes" is not
        recommended.
    precision: np.float32 or np.float64
        The level of precision to work with on internal calculations

    Returns
    -------
    result : dict
        Results dictionary containing keys: phase_index, template_index,
        orientation, correlation, and mirrored_template. phase_index is the
        phase map, with each unique integer representing a phase. template_index
        are the best matching templates for the respective phase. orientation is
        the best matching orientations expressed in Bunge convention Euler angles.
        Correlation is the matching correlation indices. mirrored template represents
        whether the original template best fits (False) or the mirror image (True).
        Each is a numpy array of shape (scan_y, scan_x, n_best) except orientation
        is of shape (scan_y, scan_x, n_best, 3).
    phase_key_dict: dictionary
        A small dictionary to translate the integers in the phase_index array
        to phase names in the original template library.


    Notes
    -----
    It is possible to run the indexation using a subset of the templates. This
    two-stage procedure is controlled through `n_keep` or `frac_keep`. If one
    of these parameters is set, the azimuthally integrated patterns are
    compared to azimuthally integrated templates in a first stage, which is
    very fast. The top matching patterns are passed to a second stage of
    full matching, whereby the in-plane angle is determined. Setting these
    parameters can usually achieve the same answer faster, but it is also
    possible an incorrect match is found.
    """
    if target == "gpu":
        # an error will be raised if cupy is not available
        if not CUPY_INSTALLED:
            raise ValueError(
                "There must be a CUDA enabled GPU and cupy must be installed."
            )
        dispatcher = cp
    else:
        dispatcher = np

    # get the dataset as a dask array
    data = _get_dask_array(signal)
    # check if we have a 4D dataset, and if not, make it
    navdim = signal.axes_manager.navigation_dimension
    if navdim == 0:
        # we assume we have a single image
        data = data[np.newaxis, np.newaxis, ...]
    elif navdim == 1:
        # we assume we have a line of images with the first dimension the line
        data = data[np.newaxis, ...]
    elif navdim == 2:
        # correct dimensions
        pass
    else:
        raise ValueError(f"Dataset has {navdim} navigation dimensions, max " "is 2")
    # change the chunking of the dataset if necessary
    if chunks is None:
        pass
    elif chunks == "auto":
        data = data.rechunk({0: "auto", 1: "auto", 2: None, 3: None})
    else:
        data = data.rechunk(chunks)

    # calculate the dimensions of the polar transform
    output_shape = get_polar_pattern_shape(
        data.shape[-2:], delta_r, delta_theta, max_r=max_r
    )
    theta_dim, r_dim = output_shape
    max_radius = r_dim * delta_r
    center = (data.shape[-2] / 2, data.shape[-1] / 2)
    # apply the intensity transform function to the images
    if intensity_transform_function is not None:
        data = data.map_blocks(intensity_transform_function)
    # combine the phases into a single library of templates to perform
    # indexation in a single step
    # TODO the library concatenation below is not memory efficient, some
    # lazy iterator would be better
    if phases is None:
        phases = library.keys()
    r_list = []  # list of all r arrays of each phase
    theta_list = []  # list of all theta arrays of each phase
    intensity_list = []  # list of all intensity arrays of each phase
    phase_index = []  # array to indicate the phase index of each template
    original_index = []  # index of the template in that phase library
    phase_key_dict = {}  # mapping the phase index to a phase name
    maximum_spot_number = 0  # to know the number of columns in template arrays
    total_template_number = 0  # to know number of rows in template arrays
    for index, phase_key in enumerate(phases):
        phase_library = library[phase_key]
        positions, intensities = _simulations_to_arrays(
            phase_library["simulations"], max_radius
        )
        r, theta = _cartesian_positions_to_polar(
            positions[:, 0], positions[:, 1], delta_r=delta_r, delta_theta=delta_theta
        )
        # ensure we don't have any out of bounds which could occur from rounding
        condition = r >= r_dim
        r[condition] = r_dim - 1
        theta[condition] = 0
        intensities[condition] = 0.0
        r_list.append(r)
        theta_list.append(theta)
        intensity_list.append(intensities.astype(precision))
        # for reconstructing the appropriate phase index and template number later
        phase_index.append(np.full((r.shape[0]), index, dtype=np.int8))
        original_index.append(np.arange(r.shape[0]))
        phase_key_dict[index] = phase_key
        # update number of spots
        if r.shape[1] > maximum_spot_number:
            maximum_spot_number = r.shape[1]
        # update number of templates
        total_template_number += r.shape[0]
    # allocate memory and concatenate arrays in list
    r = np.zeros((total_template_number, maximum_spot_number), dtype=np.int32)
    theta = np.zeros((total_template_number, maximum_spot_number), dtype=np.int32)
    intensities = np.zeros(
        (total_template_number, maximum_spot_number), dtype=precision
    )
    position = 0
    for rr, tt, ii in zip(r_list, theta_list, intensity_list):
        r[position : position + rr.shape[0], : rr.shape[1]] = rr
        theta[position : position + rr.shape[0], : rr.shape[1]] = tt
        intensities[position : position + rr.shape[0], : rr.shape[1]] = ii
        position += rr.shape[0]
    # phase_index and original index are 1D we can just concatenate
    phase_index = np.concatenate(phase_index)
    original_index = np.concatenate(original_index)

    if intensity_transform_function is not None:
        intensities = intensity_transform_function(intensities)
    # integrated intensity library for fast comparison
    integrated_templates = _get_integrated_polar_templates(
        r_dim, r, intensities, normalize_templates
    )
    # normalize the templates if required
    if normalize_templates:
        integrated_templates = _norm_rows(integrated_templates)
        intensities = _norm_rows(intensities)

    # put a limit on n_best
    max_n = _get_max_n(r.shape[0], n_keep, frac_keep)
    if n_best > max_n:
        n_best = max_n

    # copy relevant data to GPU memory if necessary
    if target == "gpu":
        integrated_templates = cp.asarray(integrated_templates)
        r = cp.asarray(r)
        theta = cp.asarray(theta)
        intensities = cp.asarray(intensities)
        indexation_function = _index_chunk_gpu
    else:
        indexation_function = _index_chunk

    indexation = data.map_blocks(
        indexation_function,
        center,
        max_radius,
        output_shape,
        precision,
        integrated_templates,
        r,
        theta,
        intensities,
        n_keep,
        frac_keep,
        n_best,
        normalize_images,
        dtype=precision,
        drop_axis=signal.axes_manager.signal_indices_in_array,
        chunks=(data.chunks[0], data.chunks[1], n_best, 4),
        new_axis=(2, 3),
    )

    # calculate number of workers
    if parallel_workers == "auto":
        # upper boundary if using CPU
        # only use number of physical cores, not logical
        parallel_workers = psutil.cpu_count(logical=False)
        if target == "gpu":
            # let's base it on the size of the free GPU memory and array blocksize
            # this is probably too many workers!
            max_gpu_mem = 0.6 * cp.cuda.Device().mem_info[0]
            blocksize = data.nbytes / data.npartitions
            max_workers = max_gpu_mem / blocksize
            if max_workers < parallel_workers:
                parallel_workers = max_workers

    with ProgressBar():
        res_index = indexation.compute(
            scheduler=scheduler, num_workers=parallel_workers, optimize_graph=True
        )

    # cupy retains memory on the GPU even after garbage collection
    # see https://docs.cupy.dev/en/stable/user_guide/memory.html
    # We manually clear the memory here to not give the user the impression that
    # their graphics card is going crazy
    if target == "gpu":
        cp.get_default_memory_pool().free_all_blocks()

    # wrangle data to results dictionary
    # we can't use a dask array to index into a dask array! Using a dask
    # array as indices for a numpy array computes the dask array automatically.
    result = {}
    result["phase_index"] = phase_index[res_index[:, :, :, 0].astype(np.int32)]
    result["template_index"] = original_index[res_index[:, :, :, 0].astype(np.int32)]
    # initialize orientations because we merge partially filled arrays by addition
    orients = 0
    for index, phase in phase_key_dict.items():
        oris = library[phase]["orientations"]
        phasemask = result["phase_index"] == index
        indices = (
            result["template_index"] * phasemask
        )  # everywhere false will get index 0 to ensure no out of bounds
        orimap = (
            oris[indices] * phasemask[..., np.newaxis]
        )  # everywhere false should not get any orientation
        # correct orientation maps with rescales and flips
        orimap[:, :, :, 1] = (
            orimap[:, :, :, 1] * res_index[:, :, :, 3]
        )  # multiply by the sign
        orimap[:, :, :, 2] = (
            orimap[:, :, :, 2] * res_index[:, :, :, 3]
        )  # multiply by the sign
        orimap[:, :, :, 0] = res_index[:, :, :, 2] * delta_theta * phasemask
        # add to orients, there should be no overlap on pixels or N
        orients = orients + orimap
    result["orientation"] = orients
    result["correlation"] = res_index[:, :, :, 1]
    result["mirrored_template"] = res_index[:, :, :, 3] == -1
    return result, phase_key_dict


def results_dict_to_crystal_map(
    results, phase_key_dict, diffraction_library=None, index=None
):
    """Export an indexation result from
    :func:`index_dataset_with_template_rotation` to a crystal map with
    `n_best` rotations, score, mirrors and one phase ID per data point.

    Parameters
    ----------
    results : dict
        Results dictionary obtained from
        :func:`index_dataset_with_template_rotation`.
    phase_key_dict : dict
        Dictionary mapping phase ID to phase name, obtained from
        :func:`index_dataset_with_template_rotation`.
    diffraction_library : diffsims.libraries.DiffractionLibrary, optional
        Used for the structures to be passed to
        :class:`orix.crystal_map.PhaseList`.
    index : int, optional
        Which of the `n_best` solutions (0-indexed) obtained from
        :func:`index_dataset_with_template_rotation` to get a crystal
        map from. Highest allowed value is `n_best` - 1. If not given,
        all solutions are used if `n_best` was more than one and
        `results["phase_index"]` only has one phase, otherwise, only the
        best solution is used.

    Returns
    -------
    orix.crystal_map.CrystalMap
        Crystal map containing `results`. The map has multiple rotations
        and properties ("correlation", "mirrored_template",
        "template_index") per point only if `n_best` passed to
        :func:`index_dataset_with_template_rotation` was more than one
        and "phase_index" only has one phase.

    Notes
    -----
    Phase's :attr:`~orix.crystal_map.Phase.point_group` must be set
    manually to the correct :class:`~orix.quaternion.Symmetry` after
    the crystal map is returned.

    Examples
    --------
    After getting `results` and `phase_key_dict` from template matching

    >>> xmap = results_dict_to_crystal_map(results, phase_key_dict)  # doctest: +SKIP
    >>> xmap.plot()  # Phase map  # doctest: +SKIP

    Getting the second best match if `n_best` passed to the template
    matching function is greater than one

    >>> xmap2 = results_dict_to_crystal_map(
    ...     results, phase_key_dict, index=1
    ... )  # doctest: +SKIP
    """
    ny, nx, n_best = results["phase_index"].shape
    if index is not None and index > n_best - 1:
        raise ValueError(f"`index` cannot be higher than {n_best - 1} (`n_best` - 1)")

    n_points = nx * ny

    # Phase ID (only one per point is allowed, always)
    if index is None:
        phase_id = results["phase_index"][:, :, 0].ravel()
    else:
        phase_id = results["phase_index"][:, :, index].ravel()
    n_phases = np.unique(phase_id).size

    x, y = np.indices((nx, ny)).reshape((2, n_points))

    if diffraction_library is not None:
        structures = diffraction_library.structures
    else:
        structures = None
    phase_list = PhaseList(names=phase_key_dict, structures=structures)

    euler = np.deg2rad(results["orientation"].reshape((n_points, n_best, 3)))
    if index is None and n_phases > 1:
        euler = euler[:, 0]  # Best match only
    elif index is not None:
        euler = euler[:, index]  # Desired match only
    euler = euler.squeeze()  # Remove singleton dimensions
    rotations = Rotation.from_euler(euler)

    props = {}
    for key in ("correlation", "mirrored_template", "template_index"):
        try:
            prop = results[key]
        except KeyError:
            warnings.warn(f"Property '{key}' was expected but not found in `results`")
            continue

        if index is None and n_phases > 1:
            prop = prop[:, :, 0].ravel()  # Best match only
        elif index is not None:
            prop = prop[:, :, index].ravel()  # Desired match only
        else:
            prop = prop.reshape((n_points, n_best))  # All
        props[key] = prop.squeeze()  # Remove singleton dimensions

    return CrystalMap(
        rotations=rotations,
        phase_id=phase_id,
        x=x,
        y=y,
        phase_list=phase_list,
        prop=props,
    )


def structure2dict(structure):
    atoms = structure.tolist()
    elements = [a.element for a in atoms]
    positions = structure.xyz
    lattice = [
        structure.lattice.a,
        structure.lattice.b,
        structure.lattice.c,
        structure.lattice.alpha,
        structure.lattice.beta,
        structure.lattice.gamma,
    ]
    return dict(
        elements=elements, positions=positions, lattice=lattice, title=structure.title
    )


def dict2structure(elements, positions, lattice, title):
    atoms = [Atom(atype=e, xyz=p) for e, p in zip(elements, positions)]
    lat = Lattice(*lattice)
    return Structure(atoms, lattice=lat, title=title)


def phase2dict(phase):
    structure_dict = structure2dict(phase.structure)
    return dict(
        name=phase.name,
        space_group=phase.space_group.number,
        point_group=phase.point_group.name,
        color=phase.color,
        structure=structure_dict,
    )


def dict2phase(name, space_group, point_group, color, structure):
    struct = dict2structure(**structure)
    return Phase(
        name=name,
        space_group=space_group,
        point_group=point_group,
        color=color,
        structure=struct,
    )



import copy
import math
import numpy as np
import scipy.optimize as opt
from matplotlib.colors import hsv_to_rgb
from hyperspy.signals import Signal2D


def _make_bivariate_histogram(
    x_position, y_position, histogram_range=None, masked=None, bins=200, spatial_std=3
):
    s0_flat = x_position.flatten()
    s1_flat = y_position.flatten()

    if masked is not None:
        temp_s0_flat = []
        temp_s1_flat = []
        for data0, data1, masked_value in zip(s0_flat, s1_flat, masked.flatten()):
            if not masked_value:
                temp_s0_flat.append(data0)
                temp_s1_flat.append(data1)
        s0_flat = np.array(temp_s0_flat)
        s1_flat = np.array(temp_s1_flat)

    if histogram_range is None:
        if s0_flat.std() > s1_flat.std():
            s0_range = (
                s0_flat.mean() - s0_flat.std() * spatial_std,
                s0_flat.mean() + s0_flat.std() * spatial_std,
            )
            s1_range = (
                s1_flat.mean() - s0_flat.std() * spatial_std,
                s1_flat.mean() + s0_flat.std() * spatial_std,
            )
        else:
            s0_range = (
                s0_flat.mean() - s1_flat.std() * spatial_std,
                s0_flat.mean() + s1_flat.std() * spatial_std,
            )
            s1_range = (
                s1_flat.mean() - s1_flat.std() * spatial_std,
                s1_flat.mean() + s1_flat.std() * spatial_std,
            )
    else:
        s0_range = histogram_range
        s1_range = histogram_range

    hist2d, xedges, yedges = np.histogram2d(
        s0_flat,
        s1_flat,
        bins=bins,
        range=[[s0_range[0], s0_range[1]], [s1_range[0], s1_range[1]]],
    )

    s_hist = Signal2D(hist2d).swap_axes(0, 1)
    s_hist.axes_manager[0].offset = xedges[0]
    s_hist.axes_manager[0].scale = xedges[1] - xedges[0]
    s_hist.axes_manager[1].offset = yedges[0]
    s_hist.axes_manager[1].scale = yedges[1] - yedges[0]
    return s_hist


def _get_corner_mask(s, corner_size=0.05):
    corner_slice_list = _get_corner_slices(s, corner_size=corner_size)
    mask = np.ones_like(s.data, dtype=bool)
    for corner_slice in corner_slice_list:
        mask[corner_slice] = False

    s_mask = s._deepcopy_with_new_data(mask).T
    return s_mask


def _get_corner_slices(s, corner_size=0.05):
    """Get signal slices for the corner positions for a Signal2D.

    Returns
    -------
    corner_array : slice tuple
        Tuple with the corner slices.

    """
    if s.axes_manager.signal_dimension != 0:
        raise ValueError("s need to have 0 signal dimensions")
    if s.axes_manager.navigation_dimension != 2:
        raise ValueError("s need to have 2 navigation dimensions")
    am = s.axes_manager.navigation_axes
    a0_range = (am[0].high_index - am[0].low_index) * corner_size
    a1_range = (am[1].high_index - am[1].low_index) * corner_size
    a0_range, a1_range = int(a0_range), int(a1_range)

    corner00_slice = np.s_[: a0_range + 1, : a1_range + 1]
    corner01_slice = np.s_[: a0_range + 1, am[1].high_index - a1_range :]
    corner10_slice = np.s_[am[0].high_index - a0_range :, : a1_range + 1]
    corner11_slice = np.s_[am[0].high_index - a0_range :, am[1].high_index - a1_range :]

    return (corner00_slice, corner01_slice, corner10_slice, corner11_slice)


def _f_min(X, p):
    plane_xyz = p[0:3]
    distance = (plane_xyz * X).sum(axis=1) + p[3]
    return distance / np.linalg.norm(plane_xyz)


def _residuals(params, X):
    return _f_min(X, params)


def _magnitude_deviations(params, X):
    plane_x_xy = params[0:2]
    plane_y_xy = params[3:5]
    distance_x = (plane_x_xy * X[:, :2]).sum(axis=1) + params[2]
    distance_y = (plane_y_xy * X[:, :2]).sum(axis=1) + params[5]
    distance = np.sqrt(((distance_x - X[:, 2]) ** 2 + (distance_y - X[:, 3]) ** 2))

    return distance - np.mean(distance)


def _plane_parameters_to_image(p, xaxis, yaxis):
    """Get a plane 2D array from plane parameters.

    Assumes a regularly spaced grid.

    Parameters
    ----------
    p : list
        4 values, like [0.1, 0.1, 0.1, 0.1]
    xaxis, yaxis : array-like
        List of x- and y-positions. For example via the axes_manager.
        s.axes_manager.signal_axes[0].axis

    Returns
    -------
    plane_image : 2D NumPy array

    Example
    -------
    >>> plane = _plane_parameters_to_image([1.2, 0.2, 3.2, 0.5], range(50), range(60))

    """
    x, y = np.meshgrid(xaxis, yaxis)
    z = (-p[0] * x - p[1] * y - p[3]) / p[2]
    return z


def _get_linear_plane_from_signal2d(signal, mask=None, initial_values=None):
    if len(signal.axes_manager.navigation_axes) != 0:
        raise ValueError("signal need to have 0 navigation dimensions")
    if len(signal.axes_manager.signal_axes) != 2:
        raise ValueError("signal need to have 2 signal dimensions")
    if initial_values is None:
        initial_values = [0.1, 0.1, 0.1, 0.1]

    sam = signal.axes_manager.signal_axes
    xaxis, yaxis = sam[0].axis, sam[1].axis
    x, y = np.meshgrid(xaxis, yaxis)
    xx, yy = x.flatten(), y.flatten()
    values = signal.data.flatten()
    points = np.stack((xx, yy, values)).T
    if mask is not None:
        if mask.__array__().shape != signal.__array__().shape:
            raise ValueError("signal and mask need to have the same shape")
        points = points[np.invert(mask).flatten()]

    p = opt.leastsq(_residuals, initial_values, args=points)[0]

    plane = _plane_parameters_to_image(p, xaxis, yaxis)
    return plane


def _get_linear_plane_by_minimizing_magnitude_variance(
    signal, mask=None, initial_values=None
):
    if len(signal.axes_manager.navigation_axes) != 2:
        raise ValueError("signal needs to have 2 navigation dimensions")
    if len(signal.axes_manager.signal_axes) != 1:
        raise ValueError("signal needs to have 1 signal dimension")
    if initial_values is None:
        initial_values = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

    signal = signal.T
    sam = signal.axes_manager.signal_axes
    xaxis, yaxis = sam[0].axis, sam[1].axis
    x, y = np.meshgrid(xaxis, yaxis)
    xx, yy = x.flatten(), y.flatten()
    values_x = signal.data[0].flatten()
    values_y = signal.data[1].flatten()
    points = np.stack((xx, yy, values_x, values_y)).T
    if mask is not None:
        if mask.__array__().shape != signal.T.__array__().shape[:2]:
            raise ValueError("signal and mask need to have the same navigation shape")
        points = points[np.invert(mask).flatten()]

    p = opt.leastsq(_magnitude_deviations, initial_values, args=points)[0]

    x, y = np.meshgrid(xaxis, yaxis)
    z_x = p[0] * x + p[1] * y + p[2]
    z_y = p[3] * x + p[4] * y + p[5]

    return np.stack((z_x, z_y), axis=-1)


def _get_limits_from_array(data, sigma=4, ignore_zeros=False, ignore_edges=False):
    if ignore_edges:
        x_lim = int(data.shape[0] * 0.05)
        y_lim = int(data.shape[1] * 0.05)
        data_array = copy.deepcopy(data[x_lim:-x_lim, y_lim:-y_lim])
    else:
        data_array = copy.deepcopy(data)
    if ignore_zeros:
        data_array = np.ma.masked_values(data_array, 0.0)
    mean = data_array.mean()
    data_variance = data_array.std() * sigma
    clim = (mean - data_variance, mean + data_variance)
    if data_array.min() > clim[0]:
        clim = list(clim)
        clim[0] = data_array.min()
        clim = tuple(clim)
    if data_array.max() < clim[1]:
        clim = list(clim)
        clim[1] = data_array.max()
        clim = tuple(clim)
    return clim


def _get_rgb_phase_array(phase, rotation=None, max_phase=2 * np.pi, phase_lim=None):
    phase = _find_phase(phase, rotation=rotation, max_phase=max_phase)
    phase = phase / (2 * np.pi)
    S = np.ones_like(phase)
    HSV = np.dstack((phase, S, S))
    RGB = hsv_to_rgb(HSV)
    return RGB


def _find_phase(phase, rotation=None, max_phase=2 * np.pi):
    if rotation is not None:
        phase = phase + math.radians(rotation)
    phase = phase % max_phase
    return phase


def _get_rgb_phase_magnitude_array(
    phase,
    magnitude,
    rotation=None,
    magnitude_limits=None,
    max_phase=2 * np.pi,
    only_phase=False,
):
    phase = _find_phase(phase, rotation=rotation, max_phase=max_phase)
    phase = phase / (2 * np.pi)

    if not only_phase:
        if magnitude_limits is not None:
            np.clip(magnitude, magnitude_limits[0], magnitude_limits[1], out=magnitude)
        magnitude_max = magnitude.max()
        if magnitude_max == 0:
            magnitude_max = 1
        magnitude = magnitude / magnitude_max
    else:
        magnitude = np.ones_like(phase)
    S = np.ones_like(phase)
    HSV = np.dstack((phase, S, magnitude))
    RGB = hsv_to_rgb(HSV)
    return RGB



from pyxem.utils import plotting as plu
from pyxem.signals import BeamShift
import hyperspy.api as hs
import matplotlib.pyplot as plt
import pytest
import numpy as np
from unittest.mock import Mock

_, ax = plt.subplots()


@pytest.fixture()
def mock_simulation():
    mock_sim = Mock()
    mock_sim.calibrated_coordinates = np.array(
        [[3, 4, 0], [5, 12, 0], [8, 15, 0], [7, 24, 0]]  # 5  # 13  # 17
    )  # 25
    mock_sim.intensities = np.array([1, 2, 3, 4])
    return mock_sim


@pytest.mark.parametrize(
    "axis, find_direct_beam, direct_beam_position, coordinate_system",
    [
        (None, False, None, "polar"),
        (ax, True, None, "cartesian"),
        (ax, False, (3, 5), "cartesian"),
    ],
)
def test_plot_sim_over_pattern(
    mock_simulation, axis, find_direct_beam, direct_beam_position, coordinate_system
):
    pattern = np.ones((15, 20))
    plu.plot_template_over_pattern(
        pattern,
        mock_simulation,
        axis,
        find_direct_beam=find_direct_beam,
        direct_beam_position=direct_beam_position,
        coordinate_system=coordinate_system,
    )


def test_plot_sim_over_pattern_fail(mock_simulation):
    pattern = np.ones((15, 20))
    with pytest.raises(NotImplementedError):
        plu.plot_template_over_pattern(
            pattern, mock_simulation, coordinate_system="dracula"
        )


def test_plot_beam_shift_color():
    s = BeamShift(np.random.random(size=(100, 100, 2)))
    plu.plot_beam_shift_color(s)
    plu.plot_beam_shift_color(
        s,
        phase_rotation=45,
        indicator_rotation=10,
        autolim=True,
        autolim_sigma=1,
        scalebar_size=10,
    )
    plu.plot_beam_shift_color(s, only_phase=True)
    plu.plot_beam_shift_color(s, autolim=False, magnitude_limits=(0, 0.5))
    fig, ax = plt.subplots()
    plu.plot_beam_shift_color(s, ax=ax)


def test_color_wheel_marker():
    color_wheel_marker = plu.make_color_wheel_marker()
    assert isinstance(color_wheel_marker, hs.plot.markers.Markers)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
from scipy.ndimage import gaussian_filter
from matplotlib import pyplot as plt


from pyxem.utils.diffraction import (
    _index_coords,
    _cart2polar,
    _polar2cart,
    remove_dead,
    find_beam_offset_cross_correlation,
    peaks_as_gvectors,
    investigate_dog_background_removal_interactive,
    find_beam_center_blur,
    find_beam_center_interpolate,
)


def test_index_coords(dp_single):
    x = np.array(
        [
            [-4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0],
            [-4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0],
            [-4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0],
            [-4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0],
            [-4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0],
            [-4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0],
            [-4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0],
            [-4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0],
        ]
    )
    y = np.array(
        [
            [-4.0, -4.0, -4.0, -4.0, -4.0, -4.0, -4.0, -4.0],
            [-3.0, -3.0, -3.0, -3.0, -3.0, -3.0, -3.0, -3.0],
            [-2.0, -2.0, -2.0, -2.0, -2.0, -2.0, -2.0, -2.0],
            [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
            [3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0],
        ]
    )
    xc, yc = _index_coords(dp_single.data)
    np.testing.assert_almost_equal(xc, x)
    np.testing.assert_almost_equal(yc, y)


def test_index_coords_non_centeral(dp_single):
    xc, yc = _index_coords(dp_single.data, origin=(0, 0))
    assert xc[0, 0] == 0
    assert yc[0, 0] == 0
    assert xc[0, 5] == 5
    assert yc[0, 5] == 0


@pytest.mark.parametrize(
    "x, y, r, theta",
    [
        (2, 2, 2.8284271247461903, -0.78539816339744828),
        (1, -2, 2.2360679774997898, 1.1071487177940904),
        (-3, 1, 3.1622776601683795, -2.81984209919315),
    ],
)
def test_cart2polar(x, y, r, theta):
    rc, thetac = _cart2polar(x=x, y=y)
    np.testing.assert_almost_equal(rc, r)
    np.testing.assert_almost_equal(thetac, theta)


@pytest.mark.parametrize(
    "r, theta, x, y",
    [
        (2.82842712, -0.78539816, 2, 2),
        (2.2360679774997898, 1.1071487177940904, 1, -2),
        (3.1622776601683795, -2.819842099193151, -3, 1),
    ],
)
def test_polar2cart(r, theta, x, y):
    xc, yc = _polar2cart(r=r, theta=theta)
    np.testing.assert_almost_equal(xc, x)
    np.testing.assert_almost_equal(yc, y)


@pytest.mark.parametrize(
    "z, center, calibration",
    [
        (
            np.array([[100, 100], [200, 200], [150, -150]]),
            np.array((127.5, 127.5)),
            0.0039,
        ),
    ],
)
def test_peaks_as_gvectors(z, center, calibration):
    gc = peaks_as_gvectors(z=z, center=center, calibration=calibration)
    ans = (z - 127.5) * calibration
    np.testing.assert_almost_equal(gc, ans)


def test_remove_dead_pixels(dp_single):
    z = dp_single.data
    dead_removed = remove_dead(z, [[3, 3]])
    assert z[3, 3] != dead_removed[3, 3]


def test_dog_background_removal_interactive(dp_single):
    """Test that this function runs without error"""
    z = dp_single
    sigma_max_list = np.arange(10, 20, 4)
    sigma_min_list = np.arange(5, 15, 6)
    investigate_dog_background_removal_interactive(z, sigma_max_list, sigma_min_list)
    plt.close("all")
    assert True


class TestCenteringAlgorithm:
    @pytest.mark.parametrize("shifts_expected", [(0, 0)])
    def test_perfectly_centered_spot(self, shifts_expected):
        z = np.zeros((50, 50))
        z[24:26, 24:26] = 1
        z = gaussian_filter(z, sigma=2, truncate=3)
        shifts = find_beam_offset_cross_correlation(z, 1, 4)
        assert np.allclose(shifts, shifts_expected, atol=0.2)

    @pytest.mark.skip(reason="FAO: M.Nord, skipping to get green for new code")
    @pytest.mark.parametrize("shifts_expected", [(+0.5, -3.5)])
    @pytest.mark.parametrize("sigma", [1, 2, 3])
    def test_single_pixel_spot(self, shifts_expected, sigma):
        z = np.zeros((50, 50))
        z[28, 24] = 1
        z = gaussian_filter(z, sigma=sigma, truncate=3)
        shifts = find_beam_offset_cross_correlation(z, 1, 6)
        assert np.allclose(shifts, shifts_expected, atol=0.2)

    @pytest.mark.parametrize("shifts_expected", [(-0.5, -4.5)])
    def test_broader_starting_square_spot(self, shifts_expected):
        z = np.zeros((50, 50))
        z[28:31, 24:27] = 1
        z = gaussian_filter(z, sigma=2, truncate=3)
        shifts = find_beam_offset_cross_correlation(z, 1, 4)
        assert np.allclose(shifts, shifts_expected, atol=0.2)


@pytest.mark.parametrize("center_expected", [(25, 29)])
@pytest.mark.parametrize("sigma", [1, 2, 3])
def test_find_beam_center_blur(center_expected, sigma):
    z = np.zeros((50, 50))
    z[28:31, 24:27] = 1
    z = gaussian_filter(z, sigma=sigma)
    shifts = find_beam_center_blur(z, 10)
    assert np.allclose(shifts, center_expected, atol=0.2)


@pytest.mark.parametrize("center_expected", [(25.97, 29.52)])
@pytest.mark.parametrize("sigma", [1, 2, 3])
def test_find_beam_center_interpolate_1(center_expected, sigma):
    z = np.zeros((50, 50))
    z[28:31, 24:28] = 1
    z = gaussian_filter(z, sigma=sigma)
    centers = find_beam_center_interpolate(z, sigma=5, upsample_factor=100, kind=3)
    assert np.allclose(centers, center_expected, atol=0.2)


@pytest.mark.parametrize("center_expected", [(44, 9)])
@pytest.mark.parametrize("sigma", [2])
def test_find_beam_center_interpolate_2(center_expected, sigma):
    """Cover unlikely case when beam is close to the edge"""
    z = np.zeros((50, 50))
    z[5:15, 41:46] = 1
    z = gaussian_filter(z, sigma=sigma)
    centers = find_beam_center_interpolate(z, sigma=5, upsample_factor=100, kind=3)
    assert np.allclose(centers, center_expected, atol=0.2)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import sys

from hyperspy.signals import Signal2D
import numpy as np
import pytest

from pyxem.generators import VirtualDarkFieldGenerator
from pyxem.signals import (
    ElectronDiffraction2D,
    DiffractionVectors,
    LearningSegment,
    VDFSegment,
    DiffractionVectors2D,
    Diffraction2D,
)


@pytest.fixture
def signal_data():
    s = ElectronDiffraction2D(np.zeros((4, 5, 6, 6)))
    s.inav[:2, :2].data[..., 0, 0] = 2
    s.inav[:2, :2].data[..., 0, 3] = 2
    s.inav[:2, :2].data[..., 3, 5] = 2

    s.inav[2:, :3].data[..., 3, 3] = 2
    s.inav[2:, :3].data[..., 3, 0] = 2

    s.inav[2, :2].data[..., 0, 0] = 1
    s.inav[2, :2].data[..., 0, 3] = 1
    s.inav[2, :2].data[..., 3, 5] = 1
    s.inav[2, :2].data[..., 3, 3] = 1
    s.inav[2, :2].data[..., 3, 0] = 1

    s.inav[:2, 2:].data[..., 5, 5] = 3
    s.inav[:2, 2:].data[..., 5, 0] = 3
    s.inav[:2, 2:].data[..., 5, 3] = 3

    s.inav[3:, 2:].data[..., 5, 5] = 3
    s.inav[3:, 2:].data[..., 5, 0] = 3
    return s


@pytest.fixture
def signal_decomposition(signal_data):
    signal_data.decomposition(algorithm="NMF", output_dimension=5, init="nndsvd")
    s_nmf = signal_data.get_decomposition_model(components=5)
    factors = s_nmf.get_decomposition_factors()
    loadings = s_nmf.get_decomposition_loadings()
    return factors, loadings


@pytest.fixture
def learning_segment(signal_decomposition):
    return LearningSegment(signal_decomposition[0], signal_decomposition[1])


class TestLearningSegment:
    def test_learning_ncc_matrix(self, learning_segment):
        ncc = learning_segment.get_ncc_matrix()
        # fmt: off
        ans = np.array(
            [
                [
                    [ 1.0,   -0.264, -0.506,  0.612, 0.0],
                    [-0.264,  1.0,   -0.401, -0.431, 0.0],
                    [-0.506, -0.401,  1.0,   -0.310, 0.0],
                    [ 0.612, -0.431, -0.310,  1.0,   0.0],
                    [ 0.0,    0.0,    0.0,    0.0,   1.0],
                ],
                [
                    [ 1.0,   -0.059, -0.073, -0.041, 0.0],
                    [-0.059,  1.0,   -0.073, -0.041, 0.0],
                    [-0.073, -0.073,  1.0,   -0.051, 0.0],
                    [-0.040, -0.041, -0.051,  1.0,   0.0],
                    [ 0.0,    0.0,    0.0,    0.0,   1.0],
                ],
            ]
        )
        # fmt: on
        assert np.allclose(ncc.data, ans, atol=1e-3)

    @pytest.mark.parametrize(
        "corr_th_factors, corr_th_loadings", [(-0.1, 0.6), (0.5, 0.5)]
    )
    def test_correlate_learning_segments(
        self, learning_segment: LearningSegment, corr_th_factors, corr_th_loadings
    ):
        learn_corr = learning_segment.correlate_learning_segments(
            corr_th_factors, corr_th_loadings
        )
        assert isinstance(learn_corr.loadings, Signal2D)
        assert isinstance(learn_corr.factors, Signal2D)

    @pytest.mark.parametrize(
        "min_intensity_threshold, min_distance,"
        "min_size, max_size, max_number_of_grains,"
        "marker_radius, threshold, exclude_border",
        [
            (0.5, 1, 1, 50, 10, 1, False, False),
            (1, 2, 4, 30, 10, 1, False, True),
            (0.1, 1, 1, 100, 100, 1, True, True),
        ],
    )
    def test_separate_learning_segments(
        self,
        learning_segment: LearningSegment,
        min_intensity_threshold,
        min_distance,
        min_size,
        max_size,
        max_number_of_grains,
        marker_radius,
        threshold,
        exclude_border,
    ):
        learn_seg = learning_segment.separate_learning_segments(
            min_intensity_threshold,
            min_distance,
            min_size,
            max_size,
            max_number_of_grains,
            marker_radius,
            threshold,
            exclude_border,
        )
        assert isinstance(learn_seg.loadings, Signal2D)
        assert isinstance(learn_seg.factors, Signal2D)


@pytest.fixture(
    params=[
        np.array(
            [
                np.array([0, 0]),
                np.array([3, 0]),
                np.array([3, 5]),
                np.array([0, 5]),
                np.array([0, 3]),
                np.array([3, 3]),
                np.array([5, 3]),
                np.array([5, 5]),
            ]
        )
    ]
)
def unique_vectors(request):
    uv = DiffractionVectors2D(request.param)
    return uv


@pytest.fixture
def vdf_segments(signal_data, unique_vectors):
    vdfgen = VirtualDarkFieldGenerator(signal_data, unique_vectors)
    vdfs = vdfgen.get_virtual_dark_field_images(radius=1)
    return vdfs.get_vdf_segments()


@pytest.fixture
def vdf_segments_cropped(vdf_segments):
    return VDFSegment(
        vdf_segments.segments.inav[:4], vdf_segments.vectors_of_segments.inav[:4]
    )


class TestVDFSegment:
    def setup_method(self):
        object1 = np.zeros((5, 5), dtype=bool)
        object1[0:2, 0:2] = True
        self.object1 = object1

        object2 = np.zeros((5, 5), dtype=bool)
        object2[3:5, 3:5] = True
        self.object2 = object2

        data = np.zeros((5, 5, 10, 10))

        data[object1, 6, 7] = 7
        data[object1, 2, 3] = 8
        data[object2, 1, 4] = 9
        data[object2, 4, 7] = 10

        self.vectors = DiffractionVectors2D([[6, 7], [2, 3], [1, 4], [4, 7]])

        self.data = Diffraction2D(data)

        self.vdf = VirtualDarkFieldGenerator(self.data, self.vectors)
        self.vdfs = self.vdf.get_virtual_dark_field_images(radius=1.0)
        self.vdf_segments = self.vdfs.get_vdf_segments()

    def test_get_vdf_ncc_matrix2(self):
        ncc = self.vdf_segments.get_ncc_matrix()
        assert ncc.data.shape == (4, 4)
        # assert that the ncc for v1 and v2 == 1 and ncc for v3 and v4 == 1
        np.testing.assert_array_almost_equal(ncc.data[:2, :2], np.ones((2, 2)))
        np.testing.assert_array_almost_equal(ncc.data[2:, 2:], np.ones((2, 2)))

    @pytest.mark.parametrize(
        "corr_threshold, vector_threshold," "segment_threshold",
        [(0.1, 1, 1), (0.9, 3, 2)],
    )
    def test_correlate_segments(
        self,
        vdf_segments: VDFSegment,
        corr_threshold,
        vector_threshold,
        segment_threshold,
    ):
        corrsegs = vdf_segments.correlate_vdf_segments(
            corr_threshold, vector_threshold, segment_threshold
        )
        assert isinstance(corrsegs.segments, Signal2D)
        assert isinstance(corrsegs.vectors_of_segments, DiffractionVectors)
        assert isinstance(corrsegs.intensities, np.ndarray)

    def test_correlate_segments_cropped(self, vdf_segments_cropped: VDFSegment):
        corrsegs = vdf_segments_cropped.correlate_vdf_segments(0.9, 1, 0)
        assert isinstance(corrsegs.segments, Signal2D)
        assert isinstance(corrsegs.vectors_of_segments, DiffractionVectors)
        assert isinstance(corrsegs.intensities, np.ndarray)

    def test_correlate_segments_small_vector_threshold(self, vdf_segments: VDFSegment):
        _ = vdf_segments.correlate_vdf_segments(
            corr_threshold=0.7, vector_threshold=0, segment_threshold=-1
        )

    def test_correlate_segments_bad_thresholds(self, vdf_segments: VDFSegment):
        with pytest.raises(
            ValueError,
            match="segment_threshold must be smaller than or equal to vector_threshold",
        ):
            _ = vdf_segments.correlate_vdf_segments(
                vector_threshold=4, segment_threshold=5
            )

    def test_get_virtual_electron_diffraction(
        self, vdf_segments: VDFSegment, signal_data
    ):
        corrsegs = vdf_segments.correlate_vdf_segments(0.1, 1, 1)
        vs = corrsegs.get_virtual_electron_diffraction(
            calibration=1, sigma=1, shape=signal_data.axes_manager.signal_shape
        )
        assert isinstance(vs, ElectronDiffraction2D)

    def test_get_virtual_electron_diffraction_no_intensities(
        self, vdf_segments: VDFSegment, signal_data
    ):
        vdf_segments.intensities = None
        with pytest.raises(
            ValueError,
            match="VDFSegment does not have the attribute intensities, required for this method",
        ):
            vdf_segments.get_virtual_electron_diffraction(
                calibration=1, sigma=1, shape=signal_data.axes_manager.signal_shape
            )

    def test_get_virtual_electron_diffraction_for_single_vectors(
        self, vdf_segments: VDFSegment, signal_data
    ):
        vs = vdf_segments.get_virtual_electron_diffraction(
            calibration=1, sigma=1, shape=signal_data.axes_manager.signal_shape
        )
        assert isinstance(vs, ElectronDiffraction2D)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

# TODO: Delete this entire module. Consider if `seperate_watershed` can be rehomed.

"""Utils for segmenting images."""

import matplotlib.pyplot as plt
import numpy as np
from numpy.ma import masked_where
from scipy.ndimage import distance_transform_edt, label, binary_erosion
from scipy.spatial import distance_matrix
from scipy.signal import convolve2d
from skimage.feature import peak_local_max
from skimage.filters import sobel, threshold_li
from skimage.morphology import disk
from skimage.segmentation import watershed
from sklearn.cluster import DBSCAN

from pyxem.utils._deprecated import deprecated


@deprecated(since="0.18.0", removal="1.0.0")
def norm_cross_corr(image, template):
    """Calculates the normalised cross-correlation between an image
    and a template at zero displacement.

    Parameters
    ----------
    image: np.array
        Image
    template: np.array
        Reference image

    Returns
    -------
    corr : float
        Normalised cross-correlation between image and template.
    """
    f, t = image - np.average(image), template - np.average(template)
    if np.all(f == 0) and np.all(t == 0):
        # both arrays contains only 0's
        corr = 1
    elif np.all(f == 0) or np.all(t == 0):
        # one arrays contains only 0's
        corr = 0
    else:
        # no divide by zero to worry about now, normal corr definition in use
        corr = np.sum(f * t) / np.sqrt(np.sum(f**2) * np.sum(t**2))

    return corr


@deprecated(since="0.18.0", removal="1.0.0")
def separate_watershed(
    vdf_temp,
    min_distance=1,
    min_size=1,
    max_size=np.inf,
    max_number_of_grains=np.inf,
    marker_radius=1,
    threshold=False,
    exclude_border=False,
    plot_on=False,
):
    """Separate segments from one VDF image using edge-detection by the
    sobel transform and the watershed segmentation implemented in
    scikit-image. See [1,2] for examples from scikit-image.

    Parameters
    ----------
    vdf_temp : np.array
        One VDF image.
    min_distance: int
        Minimum distance (in pixels) between markers for them to be
        considered separate markers for the watershed segmentation.
    min_size : float
        Grains with size (i.e. total number of pixels) below min_size
        are discarded.
    max_size : float
        Grains with size (i.e. total number of pixels) above max_size
        are discarded.
    max_number_of_grains : int
        Maximum number of grains included in the returned separated
        grains. If it is exceeded, those with highest peak intensities
        will be returned.
    marker_radius : float
        If 1 or larger, each marker for watershed is expanded to a disk
        of radius marker_radius. marker_radius should not exceed
        2*min_distance.
    threshold : bool
        If True, a mask is calculated by thresholding the VDF image by
        the Li threshold method in scikit-image. If False (default), the
        mask is the boolean VDF image.
    exclude_border : int or True, optional
        If non-zero integer, peaks within a distance of exclude_border
        from the boarder will be discarded. If True, peaks at or closer
        than min_distance of the boarder, will be discarded.
    plot_on : bool
        If True, the VDF, the mask, the distance transform
        and the separated grains will be plotted in one figure window.

    Returns
    -------
    sep : np.array
        Array containing segments from VDF images (i.e. separated
        grains). Shape: (image size x, image size y, number of grains)

    References
    ----------
    [1] https://scikit-image.org/docs/stable/auto_examples/segmentation/plot_watershed.html#sphx-glr-auto-examples-segmentation-plot-watershed-py
    [2] https://scikit-image.org/docs/stable/auto_examples/segmentation/plot_marked_watershed.html#sphx-glr-auto-examples-segmentation-plot-marked-watershed-py
    """

    # Create a mask from the input VDF image.
    if not isinstance(threshold, bool):
        mask = vdf_temp > np.max(vdf_temp) * threshold
    elif threshold:
        th = threshold_li(vdf_temp)
        mask = vdf_temp > th
    else:
        mask = vdf_temp.astype("bool")

    # Calculate the Eucledian distance from each point in the mask to the
    # nearest background point of value 0.
    distance = distance_transform_edt(mask)

    # If exclude_boarder is given, the edge of the distance is removed
    # by erosion. The distance image is used to find markers, and so the
    # erosion is done to avoid that markers are located at the edge
    # of the mask.
    if exclude_border > 0:
        distance_mask = binary_erosion(distance, structure=disk(exclude_border))
        distance = distance * distance_mask.astype("bool")

    # Find the coordinates of the local maxima of the distance transform.
    local_maxi = peak_local_max(
        distance,
        min_distance=1,
        num_peaks=max_number_of_grains,
        exclude_border=exclude_border,
        threshold_rel=None,
    )
    peak_mask = np.zeros_like(distance, dtype=bool)
    peak_mask[tuple(local_maxi.T)] = True
    local_maxi = peak_mask
    maxi_coord1 = np.where(local_maxi)

    # Discard maxima that are found at pixels that are connected to a
    # smaller number of pixels than min_size. Used as markers, these would lead
    # to segments smaller than min_size and should therefore not be
    # considered when deciding which maxima to use as markers.
    if min_size > 1:
        labels_check = label(mask)[0]
        delete_indices = []
        for i in np.arange(np.shape(maxi_coord1)[1]):
            index = np.transpose(maxi_coord1)[i]
            label_value = labels_check[index[0], index[1]]
            if len(labels_check[labels_check == label_value]) < min_size:
                delete_indices.append(i)
                local_maxi[index[0], index[1]] = False
        maxi_coord1 = np.delete(maxi_coord1, delete_indices, axis=1)

    # Cluster the maxima by DBSCAN based on min_distance. For each
    # cluster, only the maximum closest to the average maxima position is
    # used as a marker.
    if min_distance > 1 and np.shape(maxi_coord1)[1] > 1:
        clusters = DBSCAN(
            eps=min_distance,
            metric="euclidean",
            min_samples=1,
        ).fit(np.transpose(maxi_coord1))
        local_maxi = np.zeros_like(local_maxi)
        for n in np.arange(clusters.labels_.max() + 1):
            maxi_coord1_n = np.transpose(maxi_coord1)[clusters.labels_ == n]
            com = np.average(maxi_coord1_n, axis=0).astype("int")
            index = distance_matrix([com], maxi_coord1_n).argmin()
            index = maxi_coord1_n[index]
            local_maxi[index[0], index[1]] = True

    # Use the resulting maxima as markers. Each marker should have a
    # unique label value. For each maximum, generate markers with the same
    # label value in a radius given by marker_radius centered at the
    # maximum position. This is done to make the segmentation more robust
    # to local changes in pixel values around the marker.
    markers = label(local_maxi)[0]
    if marker_radius >= 1:
        disk_mask = disk(marker_radius)
        for mm in np.arange(1, np.max(markers) + 1):
            im = np.zeros_like(markers)
            im[np.where(markers == mm)] = markers[np.where(markers == mm)]
            markers_temp = convolve2d(
                im, disk_mask, boundary="fill", mode="same", fillvalue=0
            )
            markers[np.where(markers_temp)] = mm
    markers = markers * mask

    # Find the edges of the VDF image using the Sobel transform.
    elevation = sobel(vdf_temp)

    # 'Flood' the elevation (i.e. edge) image from basins at the marker
    # positions. Find the locations where different basins meet, i.e.
    # the watershed lines (segment boundaries). Only search for segments
    # (labels) in the area defined by mask.
    labels = watershed(elevation, markers=markers, mask=mask)

    sep = np.zeros(
        (np.shape(vdf_temp)[0], np.shape(vdf_temp)[1], (np.max(labels))), dtype="int32"
    )
    n, i = 1, 0
    while (np.max(labels)) > n - 1:
        sep_temp = labels * (labels == n) / n
        sep_temp = np.nan_to_num(sep_temp)
        # Discard a segment if it is too small or too large, or else add
        # it to the list of separated segments.
        if (np.sum(sep_temp, axis=(0, 1)) < min_size) or np.sum(
            sep_temp, axis=(0, 1)
        ) > max_size:
            sep = np.delete(sep, ((n - i) - 1), axis=2)
            i = i + 1
        else:
            sep[:, :, (n - i) - 1] = sep_temp
        n = n + 1
    # Put the intensity from the input VDF image into each segmented area.
    vdf_sep = np.broadcast_to(vdf_temp.T, np.shape(sep.T)) * (sep.T == 1)

    if plot_on:  # pragma: no cover
        # If segments have been discarded, make new labels that do not
        # include the discarded segments.
        if np.max(labels) != (np.shape(sep)[2]) and (np.shape(sep)[2] != 0):
            labels = sep[:, :, 0]
            for i in range(1, np.shape(sep)[2]):
                labels = labels + sep[..., i] * (i + 1)
        # If no separated particles were found, set all elements in
        # labels to 0.
        elif np.shape(sep)[2] == 0:
            labels = np.zeros(np.shape(labels))

        seps_img_sum = np.zeros_like(vdf_temp).astype("float64")
        for lbl, vdf in zip(np.arange(1, np.max(labels) + 1), vdf_sep):
            mask_l = np.zeros_like(labels).astype("bool")
            _idx = np.where(labels == lbl)
            mask_l[_idx] = 1
            seps_img_sum += vdf_temp * mask_l / np.max(vdf_temp[_idx])
            seps_img_sum[_idx] += lbl

        maxi_coord = np.where(local_maxi)

        fig, axes = plt.subplots(2, 3, sharex=True, sharey=True)
        ax = axes.ravel()

        ax[0].imshow(vdf_temp, cmap=plt.cm.magma_r)
        ax[0].axis("off")
        ax[0].set_title("VDF")

        ax[1].imshow(mask, cmap=plt.cm.gray_r)
        ax[1].axis("off")
        ax[1].set_title("Mask")

        ax[2].imshow(distance, cmap=plt.cm.gray_r)
        ax[2].axis("off")
        ax[2].set_title("Distance and markers")
        ax[2].imshow(masked_where(markers == 0, markers), cmap=plt.cm.gist_rainbow)
        ax[2].plot(maxi_coord1[1], maxi_coord1[0], "k+")
        ax[2].plot(maxi_coord[1], maxi_coord[0], "gx")

        ax[3].imshow(elevation, cmap=plt.cm.magma_r)
        ax[3].axis("off")
        ax[3].set_title("Elevation")

        ax[4].imshow(labels, cmap=plt.cm.gnuplot2_r)
        ax[4].axis("off")
        ax[4].set_title("Labels")

        ax[5].imshow(seps_img_sum, cmap=plt.cm.magma_r)
        ax[5].axis("off")
        ax[5].set_title("Segments")

    return vdf_sep


@deprecated(since="0.18.0", removal="1.0.0")
def get_gaussian2d(a, xo, yo, x, y, sigma):
    """Obtain a 2D Gaussian of amplitude a and standard deviation
    sigma, centred at (xo, yo), on a grid given by x and y.

    Parameters
    ----------
    a : float
        Amplitude. The Gaussian is simply multiplied by a.
    xo : float
        Center of Gaussian on x-axis.
    yo : float
        Center of Gaussian on y-axis.
    x : array
        Array representing the row indices of the grid the Gaussian
        should be placed on (x-axis).
    y : array
        Array representing the column indices of the grid the Gaussian
        should be placed on (y-axis).
    sigma : float
        Standard deviation of Gaussian.

    Returns
    -------
    gaussian : array
        Array with the 2D Gaussian.
    """

    # TODO This function should be removed in view of its duplication within diffsims
    gaussian = (
        a
        / (2 * np.pi * sigma**2)
        * np.exp(-((x - xo) ** 2 + (y - yo) ** 2) / (2 * sigma**2))
    )

    return gaussian



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


from pyxem.data import (
    pdnip_glass,
    zrnb_precipitate,
    au_grating,
    twinned_nanowire,
    sample_with_g,
    mgo_nanocrystals,
    tilt_boundary_data,
    simulated_strain,
)
from pyxem.data._data import Dataset
import pytest


class TestLoadData:
    @pytest.mark.skip(reason="Downloading large file takes too long")
    def test_load_pdnip(self):  # pragma: no cover
        s = pdnip_glass(allow_download=True)
        assert s.axes_manager.signal_shape == (128, 128)

    @pytest.mark.skip(reason="Downloading large file takes too long")
    def test_load_zrnb_precipitate(self):  # pragma: no cover
        s = zrnb_precipitate(allow_download=True)
        assert s.axes_manager.signal_shape == (256, 256)

    def test_load_au_grating(self):
        s = au_grating(allow_download=True)
        assert s.axes_manager.signal_shape == (254, 254)

    @pytest.mark.skip(reason="Downloading large file takes too long")
    def test_load_twined_nanowire(self):  # pragma: no cover
        s = twinned_nanowire(allow_download=True)
        assert s.axes_manager.signal_shape == (144, 144)

    @pytest.mark.skip(reason="Downloading large file takes too long")
    def test_sample_with_g(self):  # pragma: no cover
        s = sample_with_g(allow_download=True)
        assert s.axes_manager.signal_shape == (256, 256)

    @pytest.mark.skip(reason="Downloading large file takes too long")
    def test_mgo(self):  # pragma: no cover
        s = mgo_nanocrystals(allow_download=True)
        assert s.axes_manager.signal_shape == (144, 144)

    def test_simulated_tilt(self):
        tilt = tilt_boundary_data()
        assert tilt.axes_manager.signal_shape == (256, 256)
        assert tilt.axes_manager.navigation_shape == (10, 10)

    def test_simulated_tilt_pivot_point(self):
        tilt = tilt_boundary_data(correct_pivot_point=False)
        assert tilt.axes_manager.signal_shape == (256, 256)
        assert tilt.axes_manager.navigation_shape == (10, 10)

    def test_Dataset_url(self):
        pdnip = Dataset("PdNiP.zspy")
        assert "PdNiP.zspy" in pdnip.url

    def test_simulated_strain(self):
        s = simulated_strain()
        assert s.axes_manager.signal_shape == (512, 512)
        assert s.axes_manager.navigation_shape == (32, 32)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


from fractions import Fraction as frac
import numpy as np

from hyperspy.signals import Signal2D
from hyperspy._signals.lazy import LazySignal

from pyxem.utils._correlations import (
    _get_interpolation_matrix,
    _symmetry_stem,
    _corr_to_power,
)
from pyxem.signals.common_diffraction import CommonDiffraction


class Correlation2D(Signal2D, CommonDiffraction):
    """Signal class for Correlation data after applying some correlation along some axis."""

    _signal_type = "correlation"

    def get_angular_power(self, inplace=False, **kwargs):
        """Returns the power spectrum of the angular auto-correlation function
        in the form of a Signal2D class.

        This gives the fourier decomposition of the radial correlation. Due to
        Nyquist sampling the number of fourier coefficients will be equal to the
        angular range.

        Parameters
        ----------
        mask: Numpy array or Signal2D
            A bool mask of values to ignore of shape equal to the signal shape.  If the mask
            is a BaseSignal than it is iterated with the polar signal
         normalize: bool
             Normalize the radial correlation by the average value at some radius.
        inplace: bool
            From :meth:`~hyperspy.signal.BaseSignal.map`. inplace=True means the signal is
            overwritten.

        Returns
        -------
        power: Signal2D
            The power spectrum of the Signal2D
        """
        power = self.map(
            _corr_to_power, silence_warnings=True, inplace=inplace, **kwargs
        )

        s = self if inplace else power

        s.set_signal_type("power")
        fourier_axis = s.axes_manager.signal_axes[0]

        fourier_axis.name = "Fourier Coefficient"
        fourier_axis.units = "a.u"
        fourier_axis.scale = 1
        return power

    def get_summed_angular_power(self, inplace=False, **kwargs):
        """Returns the power spectrum of the summed angular auto-correlation function
        over all real space positions.  Averages the angular correlation.

        Parameters
        ----------
        inplace: bool
            From :meth:`~hyperspy.signal.BaseSignal.map`. inplace=True means the signal is
            overwritten.

        Returns
        -------
        power: Power2D
            The power spectrum of summed angular correlation
        """
        power = self.nansum().map(_corr_to_power, inplace=inplace, **kwargs)

        s = self if inplace else power

        s.set_signal_type("power")
        fourier_axis = s.axes_manager.signal_axes[0]

        fourier_axis.name = "Fourier Coefficient"
        fourier_axis.units = "a.u"
        fourier_axis.scale = 1
        return power

    def get_symmetry_coefficient(
        self,
        symmetries=[2, 3, 4, 5, 6, 7, 8, 9, 10],
        angular_range=0,
        method="average",
        include_duplicates=False,
        normalize=True,
        **kwargs
    ):
        """This function is for finding and extracting information about clusters
        based on the angular symmetries. This a pretty catch all method which has
        a couple of different operating principles.
        If k_range=None the function uses a 3 dimensional version of blob finding and attempts to
        find blobs in the 3 dimensional space, x,y,k for each symmetry.
        Parameters
        ------------
        symmetries: list
            The symmetries to calculate
        method: str
            One of max or average
        include_duplicates: bool
            Include duplicates like 2 and 4
        :return:
        """
        angles = [set(frac(j, i) for j in range(0, i)) for i in symmetries]
        if not include_duplicates:
            already_used = set()
            new_angles = []
            for a in angles:
                new_angles.append(a.difference(already_used))
                already_used = already_used.union(a)
            angles = new_angles
        num_angles = [len(a) for a in angles]
        interp = [
            _get_interpolation_matrix(
                a,
                angular_range,
                num_points=self.axes_manager.signal_axes[0].size,
                method=method,
            )
            for a in angles
        ]

        signals = self.map(
            _symmetry_stem,
            interpolation=interp,
            inplace=False,
            method=method,
            silence_warnings=True,
            **kwargs
        )
        if method in ["max", "first"]:
            normalize = False
        if normalize:
            signals.data = np.divide(signals.data, num_angles)
        # 2-D signal (x,y,k,s) for each symmetry
        signals.axes_manager.signal_axes[0].scale = 1
        signals.axes_manager.signal_axes[0].name = "Symmetry"
        signals.axes_manager.signal_axes[0].offset = symmetries[0]
        signals.axes_manager.signal_axes[0].unit = None
        return signals


class LazyCorrelation2D(LazySignal, Correlation2D):
    _lazy = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for slicing on vectors signals."""

import numpy as np
from hyperspy.signal import BaseSignal


def slice_signal(arr, col_slice, row_slice):
    if (
        hasattr(row_slice, "ndim") and row_slice.ndim == 0
    ):  # Dealing with possible 0 d array
        row_slice = [
            row_slice,
        ]
    if arr.ndim == 1:
        return arr[col_slice]
    else:
        return arr[row_slice, col_slice]


# Note this functionality should be upstreamed to hyperspy and replaced by equivalent functionality
# added to the ``isig`` attribute of hyperspy signals.
class Slicer:
    def __init__(self, signal):
        self.signal = signal

    def __getitem__(self, item):
        if isinstance(item, tuple):  # multiple dimensions
            if len(item) == 0 or len(item) > 2:
                raise ValueError(
                    "Only column and row slicing 2-D arrays is currently supported"
                )
            col_slice = self.str2slice(item[0])
            if len(item) == 2:
                row_slice = item[1]
            else:
                row_slice = slice(None)
        else:
            col_slice = self.str2slice(item)
            row_slice = slice(None)
        if self.signal.ragged:
            kwargs = dict(output_signal_size=(), output_dtype=object)
            if not isinstance(row_slice, slice):
                if not isinstance(row_slice, BaseSignal) or not row_slice.ragged:
                    raise ValueError(
                        "Only ragged boolean indexing is currently supported for ragged signals"
                    )
        else:
            kwargs = dict()

        slic = self.signal.map(
            slice_signal,
            col_slice=col_slice,
            row_slice=row_slice,
            inplace=False,
            ragged=self.signal.ragged,
            **kwargs
        )
        if self.signal.scales is not None:
            slic.scales = np.array(self.signal.scales)[col_slice]
        if self.signal.offsets is not None:
            slic.offsets = np.array(self.signal.offsets)[col_slice]
        if self.signal.column_names is not None:
            from pyxem.signals import DiffractionVectors1D, DiffractionVectors2D

            if isinstance(slic, DiffractionVectors1D) and isinstance(
                self.signal, DiffractionVectors2D
            ):
                name = np.array(self.signal.column_names)[col_slice]
                slic.axes_manager.signal_axes[0].name = "" if name is None else name
                slic.column_names = None
            else:
                slic.column_names = np.array(self.signal.column_names)[col_slice]
        if self.signal.units is not None:
            from pyxem.signals import DiffractionVectors1D, DiffractionVectors2D

            if isinstance(slic, DiffractionVectors1D) and isinstance(
                self.signal, DiffractionVectors2D
            ):
                unit = np.array(self.signal.units)[col_slice]
                slic.axes_manager.signal_axes[0].units = "" if unit is None else unit
            else:
                slic.column_names = np.array(self.signal.column_names)[col_slice]
        return slic

    def str2slice(self, item):
        if isinstance(item, str):
            item = list(self.signal.column_names).index(item)
        elif isinstance(item, (np.ndarray, list)):
            item = np.array([self.str2slice(i) for i in item])
        elif isinstance(item, (slice, int)):
            pass
        else:
            raise ValueError(
                "item must be a string or an int or an array of strings or ints"
            )
        return item



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

from pyxem.utils import cuda_utils as cutls
import dask.array as da
import numpy as np
import pytest
from unittest.mock import Mock


try:
    import cupy as cp

    CUPY_INSTALLED = True
except ImportError:
    CUPY_INSTALLED = False
    cp = np

skip_cupy = pytest.mark.skipif(not CUPY_INSTALLED, reason="cupy is required")


@skip_cupy
def test_dask_array_to_gpu():
    cutls.dask_array_to_gpu(da.array([1, 2, 3, 4]))


@skip_cupy
def test_dask_array_from_gpu():
    cutls.dask_array_from_gpu(da.array(cp.array([1, 2, 3, 4])))



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import hyperspy.api as hs
import pytest
import numpy as np

from pyxem.tests.generators.test_displacement_gradient_tensor_generator import (
    generate_test_vectors,
)
from pyxem.generators import get_DisplacementGradientMap
from pyxem.signals.strain_map import _get_rotation_matrix
from pyxem.data import simulated_strain


@pytest.fixture()
def Displacement_Grad_Map():
    xy = np.asarray([[1, 0], [0, 1]])
    deformed = hs.signals.Signal2D(generate_test_vectors(xy))
    D = get_DisplacementGradientMap(deformed, xy)
    return D


def test_rotation_matrix_formation():
    x_new = [np.random.rand(), np.random.rand()]
    R = _get_rotation_matrix(x_new)
    ratio_array = np.divide(x_new, np.matmul(R, [1, 0]))
    assert np.allclose(ratio_array[0], ratio_array[1])


def test__init__(Displacement_Grad_Map):
    strain_map = Displacement_Grad_Map.get_strain_maps()
    assert strain_map.axes_manager.navigation_size == 4


def test_signal_axes_carry_through(Displacement_Grad_Map):
    """A strain map that is calibrated, should stay calibrated when we change basis"""
    strain_map = Displacement_Grad_Map.get_strain_maps()
    strain_map.axes_manager.signal_axes[1].units = "nm"
    strain_map.axes_manager.signal_axes[0].scale = 19
    strain_alpha = strain_map.rotate_strain_basis([np.random.rand(), np.random.rand()])
    assert strain_alpha.axes_manager.signal_axes[1].units == "nm"
    assert strain_alpha.axes_manager.signal_axes[0].scale == 19


""" These are change of basis tests """


def test_something_changes(Displacement_Grad_Map):
    oneone_strain_original = Displacement_Grad_Map.get_strain_maps()
    local_D = Displacement_Grad_Map
    strain_alpha = local_D.get_strain_maps()
    oneone_strain_alpha = strain_alpha.rotate_strain_basis([1, 1])
    assert not np.allclose(
        oneone_strain_original.data, oneone_strain_alpha.data, atol=0.01
    )


def test_90_degree_rotation(Displacement_Grad_Map):
    oneone_strain_original = Displacement_Grad_Map.get_strain_maps()
    local_D = Displacement_Grad_Map
    strain_alpha = local_D.get_strain_maps()
    oneone_strain_alpha = strain_alpha.rotate_strain_basis([0, 1])
    assert np.allclose(
        oneone_strain_original.inav[2:].data,
        oneone_strain_alpha.inav[2:].data,
        atol=0.01,
    )
    assert np.allclose(
        oneone_strain_original.inav[0].data, oneone_strain_alpha.inav[1].data, atol=0.01
    )
    assert np.allclose(
        oneone_strain_original.inav[1].data, oneone_strain_alpha.inav[0].data, atol=0.01
    )


def test_going_back_and_forward_between_bases(Displacement_Grad_Map):
    """Checks that going via an intermediate strain map doesn't give incorrect answers"""
    strain_original = Displacement_Grad_Map.get_strain_maps()
    local_D = Displacement_Grad_Map
    temp_strain = local_D.get_strain_maps()
    temp_strain = temp_strain.rotate_strain_basis([np.random.rand(), np.random.rand()])
    fixed_xnew = [3.1, 4.1]
    alpha = strain_original.rotate_strain_basis(fixed_xnew)
    beta = temp_strain.rotate_strain_basis(fixed_xnew)
    assert np.allclose(alpha, beta, atol=0.01)


def test_rotation(Displacement_Grad_Map):
    """
    We should always measure the same rotations, regardless of basis
    """
    local_D = Displacement_Grad_Map
    original = local_D.get_strain_maps()
    rotation_alpha = original.rotate_strain_basis([np.random.rand(), np.random.rand()])
    rotation_beta = original.rotate_strain_basis([np.random.rand(), -np.random.rand()])

    # check the functionality has left invarient quantities invarient
    np.testing.assert_almost_equal(
        original.inav[3].data, rotation_alpha.inav[3].data, decimal=2
    )  # rotations
    np.testing.assert_almost_equal(
        original.inav[3].data, rotation_beta.inav[3].data, decimal=2
    )  # rotations


def test_trace(Displacement_Grad_Map):
    """
    Basis does affect strain measurement, but we can simply calculate suitable invariants.
    See https://en.wikipedia.org/wiki/Infinitesimal_strain_theory for details.
    """

    local_D = Displacement_Grad_Map
    original = local_D.get_strain_maps()
    rotation_alpha = original.rotate_strain_basis([1.3, +1.9])
    rotation_beta = original.rotate_strain_basis([1.7, -0.3])

    np.testing.assert_almost_equal(
        np.add(original.inav[0].data, original.inav[1].data),
        np.add(rotation_alpha.inav[0].data, rotation_alpha.inav[1].data),
        decimal=2,
    )
    np.testing.assert_almost_equal(
        np.add(original.inav[0].data, original.inav[1].data),
        np.add(rotation_beta.inav[0].data, rotation_beta.inav[1].data),
        decimal=2,
    )


@pytest.mark.parametrize("lazy", [True, False])
def test_filter_basis(lazy):
    """
    This test is a workflow test for filtering basis for strain mapping.
    """
    signal = simulated_strain(
        navigation_shape=(16, 16),
        signal_shape=(128, 128),
        disk_radius=10,
        num_electrons=None,
        strain_matrix=None,
        lazy=True,
    )

    temp = signal.template_match_disk(disk_r=10, subtract_min=False)
    vectors = temp.get_diffraction_vectors(threshold_abs=0.4, min_distance=5)
    vectors = vectors.filter_magnitude(0.1, 1)
    unstrained_vectors = vectors.inav[0, 0]
    filtered_vectors = vectors.filter_basis(unstrained_vectors)
    filtered_vectors.compute()
    assert filtered_vectors.data.shape == (16, 16, 4, 3)
    assert filtered_vectors.axes_manager[0].name == "x"
    assert filtered_vectors.axes_manager[1].name == "y"


@pytest.mark.parametrize("lazy", [True, False])
def test_get_strain(lazy):
    """
    This test is a workflow test for the strain map. It checks that the strain map can be
    calculated and that the results are as expected.
    """
    signal = simulated_strain(
        navigation_shape=(16, 16),
        signal_shape=(128, 128),
        disk_radius=10,
        num_electrons=None,
        strain_matrix=None,
        lazy=True,
    )

    temp = signal.template_match_disk(disk_r=10, subtract_min=False)
    vectors = temp.get_diffraction_vectors(threshold_abs=0.4, min_distance=5)
    vectors = vectors.filter_magnitude(0.1, 1)
    unstrained_vectors = vectors.inav[0, 0]
    strain = vectors.get_strain_maps(unstrained_vectors=unstrained_vectors)

    assert strain.axes_manager[1].name == signal.axes_manager.navigation_axes[0].name
    assert strain.axes_manager[1].scale == signal.axes_manager.navigation_axes[0].scale
    assert strain.axes_manager[1].units == signal.axes_manager.navigation_axes[0].units
    assert strain.axes_manager[2].name == signal.axes_manager.navigation_axes[1].name
    assert strain.axes_manager[2].scale == signal.axes_manager.navigation_axes[1].scale
    assert strain.axes_manager[2].units == signal.axes_manager.navigation_axes[1].units


def test_get_strain_error():
    signal = simulated_strain(
        navigation_shape=(16, 16),
        signal_shape=(128, 128),
        disk_radius=10,
        num_electrons=None,
        strain_matrix=None,
        lazy=True,
    )

    temp = signal.template_match_disk(disk_r=10, subtract_min=False)
    vectors = temp.get_diffraction_vectors(threshold_abs=0.4, min_distance=5)
    vectors = vectors.filter_magnitude(0.1, 0.2)
    unstrained_vectors = vectors.inav[0, 0]
    with pytest.raises(ValueError):
        vectors.get_strain_maps(
            unstrained_vectors=unstrained_vectors, return_residuals=True
        )



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from hyperspy.signals import Signal2D
from hyperspy.roi import Line2DROI
from diffsims.utils.ring_pattern_utils import generate_ring_pattern

from pyxem.signals import ElectronDiffraction2D
from pyxem.generators import CalibrationGenerator
from pyxem.libraries import CalibrationDataLibrary


@pytest.fixture
def input_parameters():
    x0 = [95, 1200, 2.8, 450, 1.5, 10]
    return x0


@pytest.fixture
def affine_answer():
    a = np.asarray(
        [[1.06651526, 0.10258988, 0.0], [0.10258988, 1.15822961, 0.0], [0.0, 0.0, 1.0]]
    )
    return a


@pytest.fixture
def ring_pattern(input_parameters):
    x0 = input_parameters
    ring_data = generate_ring_pattern(
        image_size=256,
        mask=True,
        mask_radius=10,
        scale=x0[0],
        amplitude=x0[1],
        spread=x0[2],
        direct_beam_amplitude=x0[3],
        asymmetry=x0[4],
        rotation=x0[5],
    )

    return ElectronDiffraction2D(ring_data)


@pytest.fixture
def grating_image(request):
    #  Create a dummy X-grating image
    data = np.zeros((200, 200))
    data[:, 10:20] = 100
    data[:, 30:40] = 50
    data[:, 150:160] = 50
    data[:, 170:180] = 100
    im = Signal2D(data)
    return im


@pytest.fixture
def calgen(request, ring_pattern, grating_image):
    return CalibrationGenerator(
        diffraction_pattern=ring_pattern, grating_image=grating_image
    )


@pytest.fixture
def cal_dist(request, calgen):
    calgen.get_elliptical_distortion(
        mask_radius=10,
        direct_beam_amplitude=450,
        scale=95,
        amplitude=1200,
        asymmetry=1.5,
        spread=2.8,
        rotation=10,
    )
    return calgen


class TestCalibrationGenerator:
    def test_init(self, calgen):
        assert isinstance(calgen.diffraction_pattern, ElectronDiffraction2D)

    def test_str(self, calgen):
        print(calgen)

    def test_get_elliptical_distortion(self, cal_dist, input_parameters, affine_answer):
        np.testing.assert_allclose(cal_dist.affine_matrix, affine_answer, rtol=1e-3)
        np.testing.assert_allclose(cal_dist.ring_params, input_parameters, rtol=1e-3)

    def test_get_distortion_residuals(self, cal_dist):
        residuals = cal_dist.get_distortion_residuals(mask_radius=10, spread=2)
        assert isinstance(residuals, ElectronDiffraction2D)

    def test_plot_corrected_diffraction_pattern(self, cal_dist):
        cal_dist.plot_corrected_diffraction_pattern()

    def test_get_diffraction_calibration(self, cal_dist):
        cal_dist.get_diffraction_calibration(mask_length=30, linewidth=5)
        np.testing.assert_almost_equal(
            cal_dist.diffraction_calibration, 0.01061096, decimal=3
        )

    def test_get_navigation_calibration(self, calgen):
        line = Line2DROI(x1=2.5, y1=13.0, x2=193.0, y2=12.5, linewidth=3.5)
        value = calgen.get_navigation_calibration(
            line_roi=line, x1=12.0, x2=172.0, n=1, xspace=500.0
        )
        np.testing.assert_almost_equal(calgen.navigation_calibration, value, decimal=3)

    def test_get_rotation_calibration(self, calgen):
        real_line = Line2DROI(
            x1=2.69824, y1=81.4867, x2=229.155, y2=61.6898, linewidth=3
        )
        recip_line = Line2DROI(
            x1=-0.30367, y1=-1.21457, x2=0.344978, y2=1.24927, linewidth=0.115582
        )
        value = calgen.get_rotation_calibration(
            real_line=real_line, reciprocal_line=recip_line
        )
        np.testing.assert_almost_equal(value, -80.24669411537899, decimal=3)

    def test_plot_calibrated_data_dp(self, cal_dist):
        cal_dist.get_diffraction_calibration(mask_length=30, linewidth=5)
        cal_dist.plot_calibrated_data(data_to_plot="au_x_grating_dp")

    def test_plot_calibrated_data_im(self, calgen):
        line = Line2DROI(x1=2.5, y1=13.0, x2=193.0, y2=12.5, linewidth=3.5)
        calgen.get_navigation_calibration(
            line_roi=line, x1=12.0, x2=172.0, n=1, xspace=500.0
        )
        calgen.plot_calibrated_data(data_to_plot="au_x_grating_im")


class TestGetCorrectionMatrix:
    def test_get_correction_rotation_only(self, calgen):
        calgen.rotation_angle = -80.24669411537899
        corr = calgen.get_correction_matrix()
        np.testing.assert_almost_equal(
            corr,
            np.array(
                [
                    [0.16940637, 0.98554629, 0.0],
                    [-0.98554629, 0.16940637, 0.0],
                    [0.0, 0.0, 1.0],
                ]
            ),
            decimal=3,
        )

    def test_get_correction_affine_only(self, calgen):
        affine = np.array(
            [
                [0.97579077, 0.01549655, 0.0],
                [0.01549655, 0.99008051, 0.0],
                [0.0, 0.0, 1.0],
            ]
        )
        calgen.affine_matrix = affine
        corr = calgen.get_correction_matrix()
        np.testing.assert_almost_equal(corr, affine, decimal=3)

    def test_get_correction_affine_and_rotation(self, calgen):
        affine = np.array(
            [
                [0.97579077, 0.01549655, 0.0],
                [0.01549655, 0.99008051, 0.0],
                [0.0, 0.0, 1.0],
            ]
        )
        calgen.affine_matrix = affine
        real_line = Line2DROI(
            x1=2.69824, y1=81.4867, x2=229.155, y2=61.6898, linewidth=3
        )
        recip_line = Line2DROI(
            x1=-0.30367, y1=-1.21457, x2=0.344978, y2=1.24927, linewidth=0.115582
        )
        value = calgen.get_rotation_calibration(
            real_line=real_line, reciprocal_line=recip_line
        )
        corr = calgen.get_correction_matrix()
        np.testing.assert_almost_equal(
            corr,
            np.array(
                [
                    [0.1805777, 0.9783954, 0.0],
                    [-0.9590618, 0.1524534, 0.0],
                    [0.0, 0.0, 1.0],
                ]
            ),
            decimal=3,
        )

    def test_no_attributes_correction_matrix(self, calgen):
        with pytest.raises(
            ValueError,
            match="requires either an affine matrix to correct distortion",
        ):
            calgen.get_correction_matrix()


@pytest.fixture
def empty_calibration_library(request):
    return CalibrationDataLibrary()


@pytest.fixture
def empty_calgen(request, empty_calibration_library):
    return CalibrationGenerator()


class TestEmptyCalibrationGenerator:
    def test_get_elliptical_distortion(
        self, empty_calgen, input_parameters, affine_answer
    ):
        with pytest.raises(
            ValueError,
            match="This method requires a calibration diffraction"
            " pattern to be provided. Please set"
            " self.diffraction_pattern equal to some Signal2D.",
        ):
            empty_calgen.get_elliptical_distortion(
                mask_radius=10,
                direct_beam_amplitude=450,
                scale=95,
                amplitude=1200,
                asymmetry=1.5,
                spread=2.8,
                rotation=10,
            )

    def test_get_distortion_residuals_no_data(self, empty_calgen):
        with pytest.raises(ValueError, match="requires an Au X-grating diffraction"):
            empty_calgen.get_distortion_residuals(mask_radius=10, spread=2)

    def test_get_distortion_residuals_no_affine(self, calgen):
        with pytest.raises(ValueError, match="requires a distortion matrix"):
            calgen.get_distortion_residuals(mask_radius=10, spread=2)

    def test_plot_corrected_diffraction_pattern_no_data(self, empty_calgen):
        with pytest.raises(ValueError, match="requires an Au X-grating diffraction"):
            empty_calgen.plot_corrected_diffraction_pattern()

    def test_plot_corrected_diffraction_pattern_no_affine(self, calgen):
        with pytest.raises(ValueError, match="requires a distortion matrix"):
            calgen.plot_corrected_diffraction_pattern()

    def test_get_diffraction_calibration_no_data(self, empty_calgen):
        with pytest.raises(ValueError, match="requires an Au X-grating diffraction"):
            empty_calgen.get_diffraction_calibration(mask_length=30, linewidth=5)

    def test_get_diffraction_calibration_no_affine(self, calgen):
        with pytest.raises(ValueError, match="requires a distortion matrix"):
            calgen.get_diffraction_calibration(mask_length=30, linewidth=5)

    def test_get_navigation_calibration_no_data(self, empty_calgen):
        line = Line2DROI(x1=2.5, y1=13.0, x2=193.0, y2=12.5, linewidth=3.5)

        with pytest.raises(ValueError, match="requires an Au X-grating image"):
            empty_calgen.get_navigation_calibration(
                line_roi=line, x1=12.0, x2=172.0, n=1, xspace=500.0
            )
